unit mxBNMath;

// Binuar 1.1
// Big number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{.$DEFINE BN_INLINE}

{$DEFINE KARATSUBA_V2}

{$DEFINE X64_INC_AH}
{$DEFINE LOOP_INTX}
{$IFDEF CPUX64}
  {$IFDEF VER_DXE2}
    {$UNDEF X64_INC_AH}
    {$UNDEF LOOP_INTX}
  {$ENDIF}
{$ENDIF}

interface

uses
  Windows, SysConst, SysUtils;

const
  Karatsuba_mul_TCount = 80;
  Karatsuba_sqr_TCount = 60;

type Int8   = ShortInt;
type UInt8  = Byte;
type Int16  = SmallInt;
type UInt16 = Word;
type Int32  = Integer;
type UInt32 = Cardinal;

type
  {$IFDEF CPUX64}
  PIntX = ^IntX;
  IntX = Int64;

  PUIntX = ^UIntX;
  UIntX = UInt64;

  {$IFDEF LOOP_INTX}
  LoopInt = IntX;
  LoopUInt = UIntX;
  {$ELSE}
  LoopInt = Int32;
  LoopUInt = UInt32;
  {$ENDIF}
  {$ELSE}
  PIntX = ^IntX;
  IntX = Int32;

  PUIntX = ^UIntX;
  UIntX = UInt32;

  LoopInt = IntX;
  LoopUInt = UIntX;
  {$ENDIF}

type
  P_BNX = ^T_BNX;
  T_BNX = record
    Sign   : ShortInt; // -1 0 +1
    Count  : IntX;     //    Data
                       // Number of Data array elements
    HIndex : IntX;     // HIGH(Data)
    PLData : PUIntX;   // @Data[0]
    PHData : PUIntX;   // @Data[HIndex]
    Data   : array of UIntX;
  end;

function  BN_dup (var V: T_BNX; const A: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_cdup (var V: T_BNX; const A: T_BNX; ACount: IntX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_sdup (var V: T_BNX; ASign: ShortInt; const A: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_ndup (var V: T_BNX; const A: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_hdup (var V: T_BNX; const A: T_BNX; AIndex: IntX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_shl_(var a: T_BNX; ABits: IntX); overload;
procedure BN_shl (var a: T_BNX; ABits: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_shl_(var b: T_BNX; const a: T_BNX; ABits: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_shl (var b: T_BNX; const a: T_BNX; ABits: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_shr_(var a: T_BNX; ABits: IntX); overload;
procedure BN_shr (var a: T_BNX; ABits: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_shr_(var b: T_BNX; const a: T_BNX; ABits: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_shr (var b: T_BNX; const a: T_BNX; ABits: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_cshl_(var a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_cshl (var a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_cshl_(var b: T_BNX; const a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_cshl (var b: T_BNX; const a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_cshr_(var a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_cshr (var a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_cshr_(var b: T_BNX; const a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_cshr (var b: T_BNX; const a: T_BNX; C: IntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_add (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_add_(var a1: T_BNX; const a2: T_BNX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_add_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_add_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_add_sint (var a3: T_BNX; const a1: T_BNX; a2: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_add_sint_(var a1: T_BNX; a2: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_sub (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_sub_(var a1: T_BNX; const a2: T_BNX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_sub_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX); overload;
procedure BN_sub_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_sub_sint (var a3: T_BNX; const a1: T_BNX; a2: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_sub_sint_(var a1: T_BNX; a2: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_mul_div_sign (ASign1, ASign2: ShortInt): ShortInt;

procedure BN_mul_base (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_mul_base_(var a1: T_BNX; const a2: T_BNX);
procedure BN_mul_base_nz (var a3: T_BNX; const a1, a2: T_BNX; ACount1, ACount2: IntX; ASign3: ShortInt);
procedure BN_mul_base_nz_(var a1: T_BNX; const a2: T_BNX; ACount1, ACount2: IntX; ASign3: ShortInt);
procedure BN_mul (var a3: T_BNX; const a1, a2: T_BNX);
procedure BN_mul_(var a1: T_BNX; const a2: T_BNX);
procedure BN_mul_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
procedure BN_mul_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX);
procedure BN_mul_int_nz (var a3: T_BNX; const a1: T_BNX; a2: UIntX; ACount1: IntX; ASign3: ShortInt);
procedure BN_mul_int_nz_(var a1: T_BNX; a2: UIntX; ACount1: IntX; ASign3: ShortInt);
procedure BN_mul_sint (var a3: T_BNX; const a1: T_BNX; a2: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_mul_intn (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX);
procedure BN_mul_sintn (var a3: T_BNX; const a1: T_BNX; a2: IntX; n: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_div_mod (var adiv, amod: T_BNX; var a1, a2: T_BNX): Integer;
function  BN_mod (var amod: T_BNX; var a1, a2: T_BNX): Integer;
function  BN_div (var adiv: T_BNX; var a1, a2: T_BNX): Integer; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_div_int (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_div_sint (var adiv: T_BNX; const a1: T_BNX; a2: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_div_intn (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_div_sintn (var adiv: T_BNX; const a1: T_BNX; a2: IntX; n: IntX); {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_div_mod_int (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX): UIntX;
function  BN_div_mod_sint (var adiv: T_BNX; const a1: T_BNX; a2: IntX): UIntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_div_mod_intn (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX; n: IntX): UIntX;
function  BN_div_mod_sintn (var adiv: T_BNX; const a1: T_BNX; a2: IntX; n: IntX): UIntX; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_exp_base (var P: T_BNX; const ABase: T_BNX; AExp: UIntX);
procedure BN_exp (var P: T_BNX; const ABase: T_BNX; AExp: UIntX);
procedure BN_exp_int (var P: T_BNX; ASign: ShortInt; ABase: UIntX; AExp: UIntX);
procedure BN_exp_sint (var P: T_BNX; ABase: IntX; AExp: UIntX);
procedure BN_10exp (var P: T_BNX; ASign: ShortInt; AExp: UIntX);
procedure BN_sqr_base (var P: T_BNX; const ABase: T_BNX);
procedure BN_sqr_base_(var ABase: T_BNX);
procedure BN_sqr (var P: T_BNX; const ABase: T_BNX);
procedure BN_sqr_(var ABase: T_BNX);
procedure BN_exp3 (var P: T_BNX; const ABase: T_BNX);
procedure BN_exp4 (var P: T_BNX; const ABase: T_BNX);
procedure BN_exp5 (var P: T_BNX; const ABase: T_BNX);

function  BN_root (var aroot: T_BNX; var a: T_BNX; n: UIntX): Integer;
function  BN_root_mod (var aroot, amod: T_BNX; var a: T_BNX; n: UIntX): Integer;
function  BN_sqrt (var aroot: T_BNX; var a: T_BNX): Integer;
function  BN_sqrt_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;
function  BN_root3 (var aroot: T_BNX; var a: T_BNX): Integer;
function  BN_root3_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;
function  BN_root4 (var aroot: T_BNX; var a: T_BNX): Integer;
function  BN_root4_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;
function  BN_root5 (var aroot: T_BNX; var a: T_BNX): Integer;
function  BN_root5_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;

function  BN_m2n (const a: T_BNX): UIntX;
function  BN_m2n_(const a: T_BNX; var sha: T_BNX): UIntX;
function  BN_m2n_int (a: UIntX): UIntX;
function  BN_m2n_int_(a: UIntX; var sha: UIntX): UIntX;

procedure BN_gcd (var R: T_BNX; var m, n: T_BNX);
procedure BN_gcd_bin (var R: T_BNX; var m, n: T_BNX);
procedure BN_lcm (var R: T_BNX; var m, n: T_BNX);

procedure BN_fact (var R: T_BNX; const a: T_BNX);
procedure BN_fact_int (var R: T_BNX; a: UIntX);

implementation

uses
  mxBNMathUtils, mxBNMathProcs;

{
  V = A
}
function BN_dup (var V: T_BNX; const A: T_BNX): IntX;
begin
  Result := BN_Alloc (V, A.Sign, A.Count);
  BN_CopyData (V.PLData, A.PLData, V.Count);
end;

{
  V <- A
    ACount 
  copy the first ACount elements
}
function BN_cdup (var V: T_BNX; const A: T_BNX; ACount: IntX): IntX;
var
  VCount  : IntX;
  VDCount : IntX;
begin
  Result := BN_Alloc (V, A.Sign, ACount);

  VCount := A.Count;
  VDCount := V.Count - VCount;
  if VDCount > 0 then begin
    BN_CopyData (V.PLData, A.PLData, VCount);
    BN_ZeroData (@V.Data[VCount], VDCount);
  end else begin
    BN_CopyData (V.PLData, A.PLData, V.Count);
  end;
end;

{
  |V| = |A|
  V.Sign = ASign
}
function BN_sdup (var V: T_BNX; ASign: ShortInt; const A: T_BNX): IntX;
begin
  Result := BN_Alloc (V, ASign, A.Count);
  BN_CopyData (V.PLData, A.PLData, V.Count);
end;

{
  V = -A
}
function BN_ndup (var V: T_BNX; const A: T_BNX): IntX;
begin
  Result := BN_Alloc (V, -A.Sign, A.Count);
  BN_CopyData (V.PLData, A.PLData, V.Count);
end;

{
  V <- A
  ,    #AIndex
  copy, starting from the element #AIndex
}
function BN_hdup (var V: T_BNX; const A: T_BNX; AIndex: IntX): IntX;
var
  VCount : IntX;
begin
  VCount := A.Count - AIndex;
  if (A.Sign = 0) or (VCount <= 0) then begin
    Result := BN_AssignZero (V);
    Exit;
  end;

  Result := BN_Alloc (V, A.Sign, VCount);
  BN_CopyData (V.PLData, @A.Data[AIndex], V.Count);
end;

{
  Result = a div (32|64)
  amod   = a mod (32|64)
  rmod   = (32|64) - amod
}
function BNX_div_mod_x_asm (a: UIntX; var amod: UIntX; var rmod: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a     |
// rdx | @amod | rcx
// r8  | @rmod |
//     | 64    | r9
asm
        mov     rax, rcx
        mov     rcx, rdx
        xor     rdx, rdx
        mov     r9, 64
        div     r9
        mov     [rcx], rdx
        sub     r9, rdx
        mov     [r8], r9
end;
{$ELSE}
// eax | a     |
// edx | @amod | edi
// ecx | @rmod |
//     | 32    | ebx
asm
        push    ebx
        push    edi

        mov     edi, edx
        xor     edx, edx
        mov     ebx, 32
        div     ebx
        mov     [edi], edx
        sub     ebx, edx
        mov     [ecx], ebx

        pop     edi
        pop     ebx
end;
{$ENDIF}

{
  shift left by ABits bits, ABits > 0
  a = a shl ABits
}
procedure BN_shl_(var a: T_BNX; ABits: IntX);
var
  I       : LoopInt;
  P       : IntX;
  C       : IntX;
  CB, CX  : UIntX;
  V, U    : UIntX;
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if ABits <= 0 then Exit;

  C := BNX_div_mod_x_asm (ABits, CB, CX);

  VCount := BN_DataCount (a);
  if CB = 0 then VRCount := VCount + C else VRCount := VCount + C + 1;
  if VRCount <= VCount then begin
    BN_AssignZero (a);
    Exit;
  end;

  BN_Resize (a, VRCount);

  if CB = 0 then begin
    // VRCount = VCount + C
    if C > 0 then begin
      I := VCount - 1;
      BN_CopyDataDesc (@a.Data[I + C], @a.Data[I], VCount);
      BN_ZeroData (a.PLData, C);
    end;
  end else begin
    // VRCount = VCount + C + 1
    P := VRCount - 1;
    a.Data[P] := 0;

    for I := VCount - 1 downto 0 do begin
      V := a.Data[I] shr CX;
      U := a.Data[P] or V;
      a.Data[P] := U;
      Dec (P);
      U := a.Data[I] shl CB;
      a.Data[P] := U;
    end;

    if C > 0 then begin
      BN_ZeroData (a.PLData, C);
    end;
    BN_Pack (a);
  end;
end;

{
  shift left by ABits bits
  a = a shl ABits
}
procedure BN_shl (var a: T_BNX; ABits: IntX);
begin
  if ABits = 0 then Exit;

  if ABits > 0 then begin
    BN_shl_(a, ABits);
  end else begin
    BN_shr_(a, -ABits);
  end;
end;

{
  shift left by ABits bits, ABits > 0
  b = a shl ABits
}
procedure BN_shl_(var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  BN_dup (b, a);
  if ABits <= 0 then Exit;

  BN_shl_(b, ABits);
end;

{
  shift left by ABits bits
  b = a shl ABits
}
procedure BN_shl (var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  if ABits = 0 then begin
    BN_dup (b, a);
    Exit;
  end;

  if ABits > 0 then begin
    BN_shl_(b, a, ABits);
  end else begin
    BN_shr_(b, a, -ABits);
  end;
end;

{
  shift right by ABits bits, ABits > 0
  a = a shr ABits
}
procedure BN_shr_(var a: T_BNX; ABits: IntX);
var
  I       : LoopInt;
  C       : IntX;
  CB, CX  : UIntX;
  V, U    : UIntX;
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if ABits <= 0 then Exit;

  C := BNX_div_mod_x_asm (ABits, CB, CX);

  VCount := BN_DataCount (a);
  VRCount := VCount - C;
  if VRCount <= 0 then begin
    BN_AssignZero (a);
    Exit;
  end;

  if CB = 0 then begin
    if C > 0 then begin
      BN_CopyData (a.PLData, @a.Data[C], VRCount);
    end;

    BN_Resize (a, VRCount);
  end else begin
    V := a.Data[C] shr CB;
    a.Data[0] := V;

    for I := 1 to VRCount - 1 do begin
      V := a.Data[I + C] shl CX;
      U := a.Data[I - 1] or V;
      a.Data[I - 1] := U;
      U := a.Data[I + C] shr CB;
      a.Data[I] := U;
    end;

    BN_Resize (a, VRCount);
    BN_Pack (a);
  end;
end;

{
  shift right by ABits bits
  a = a shr ABits
}
procedure BN_shr (var a: T_BNX; ABits: IntX);
begin
  if ABits <= 0 then Exit;

  if ABits > 0 then begin
    BN_shr_(a, ABits);
  end else begin
    BN_shl_(a, -ABits);
  end;
end;

{
  shift right by ABits bits, ABits > 0
  b = a shr ABits
}
procedure BN_shr_(var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  BN_dup (b, a);
  if ABits <= 0 then Exit;

  BN_shr_(b, ABits);
end;

{
  shift right by ABits bits
  b = a shr ABits
}
procedure BN_shr (var b: T_BNX; const a: T_BNX; ABits: IntX);
begin
  if ABits = 0 then begin
    BN_dup (b, a);
    Exit;
  end;

  if ABits > 0 then begin
    BN_shr_(b, a, ABits);
  end else begin
    BN_shl_(b, a, -ABits);
  end;
end;

{
  shift left by C*(32|64) bits, C > 0
  a = a shl C*(32|64)
}
procedure BN_cshl_(var a: T_BNX; C: IntX);
var
  I       : IntX;
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if C <= 0 then Exit;

  VCount := BN_DataCount (a);
  VRCount := VCount + C;
  if VRCount <= VCount then begin
    BN_AssignZero (a);
    Exit;
  end;

  BN_Resize (a, VRCount);
  I := VCount - 1;
  BN_CopyDataDesc (@a.Data[I + C], @a.Data[I], VCount);
  BN_ZeroData (a.PLData, C);
end;

{
  shift left by C*(32|64) bits
  a = a shl C*(32|64)
}
procedure BN_cshl (var a: T_BNX; C: IntX);
begin
  if C = 0 then Exit;

  if C > 0 then begin
    BN_cshl_(a, C);
  end else begin
    BN_cshr_(a, -C);
  end;
end;

{
  shift left by C*(32|64) bits, C > 0
  b = a shl C*(32|64)
}
procedure BN_cshl_(var b: T_BNX; const a: T_BNX; C: IntX);
begin
  BN_dup (b, a);
  if C <= 0 then Exit;

  BN_cshl_(b, C);
end;

{
  shift left by C*(32|64) bits
  b = a shl C*(32|64)
}
procedure BN_cshl (var b: T_BNX; const a: T_BNX; C: IntX);
begin
  if C = 0 then begin
    BN_dup (b, a);
    Exit;
  end;

  if C > 0 then begin
    BN_cshl_(b, a, C);
  end else begin
    BN_cshr_(b, a, -C);
  end;
end;

{
  shift right by C*(32|64) bits, C > 0
  a = a shr C*(32|64)
}
procedure BN_cshr_(var a: T_BNX; C: IntX);
var
  VCount  : IntX;
  VRCount : IntX; // Result count
begin
  if C <= 0 then Exit;

  VCount := BN_DataCount (a);
  VRCount := VCount - C;
  if VRCount <= 0 then begin
    BN_AssignZero (a);
    Exit;
  end;

  BN_CopyData (a.PLData, @a.Data[C], VRCount);
  BN_Resize (a, VRCount);
end;

{
  shift right by C*(32|64) bits
  a = a shr C*(32|64)
}
procedure BN_cshr (var a: T_BNX; C: IntX);
begin
  if C <= 0 then Exit;

  if C > 0 then begin
    BN_cshr_(a, C);
  end else begin
    BN_cshl_(a, -C);
  end;
end;

{
  shift right by C*(32|64) bits, C > 0
  b = a shr C*(32|64)
}
procedure BN_cshr_(var b: T_BNX; const a: T_BNX; C: IntX);
begin
  BN_dup (b, a);
  if C <= 0 then Exit;

  BN_cshr_(b, C);
end;

{
  shift right by C*(32|64) bits
  b = a shr C*(32|64)
}
procedure BN_cshr (var b: T_BNX; const a: T_BNX; C: IntX);
begin
  if C = 0 then begin
    BN_dup (b, a);
    Exit;
  end;

  if C > 0 then begin
    BN_cshr_(b, a, C);
  end else begin
    BN_cshl_(b, a, -C);
  end;
end;

type
  T_BNX_add_sub_asm_proc = function (a1, a2, a3: PUIntX; XCount, DCount: IntX): UIntX;
  T_BNX_add_sub_int_asm_proc = function (a1: PUIntX; a2: UIntX; a3: PUIntX; Count: IntX): UIntX;

{
  a3 = a1 + a2
  |a1| >= |a2|
  XCount = min (a1.Count, a2.Count) = a2.Count
  DCount = a1.Count - a2.Count
}
function BNX_add_asm (a1, a2, a3: PUIntX; XCount, DCount: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1    |
// rdx | @a2    |
// r8  | @a3    |
// r9  | XCount |
//     | DCount | r9
asm
        clc

        mov     rax, qword ptr [rcx]
     // mov     r10, qword ptr [rdx]
     // add     rax, r10
        add     rax, qword ptr [rdx]
        mov     qword ptr [r8], rax

        dec     r9
        jz      @b1
    @a1:
        lea     rdx, [rdx + 8]
        lea     rcx, [rcx + 8]
        lea     r8, [r8 + 8]
        mov     rax, qword ptr [rcx]
     // mov     r10, qword ptr [rdx]
     // adc     rax, r10
        adc     rax, qword ptr [rdx]
        mov     qword ptr [r8], rax
        dec     r9
        jnz     @a1
    @b1:
        mov     r9, [DCount]
        dec     r9
        jl      @b2
        mov     rdx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     rdx, rdx
        // *)
    @a2:
        lea     rcx, [rcx + 8]
        lea     r8, [r8 + 8]
        mov     rax, qword ptr [rcx]
        adc     rax, rdx
        mov     qword ptr [r8], rax
        dec     r9
        jge     @a2
    @b2:
        jc      @c
        xor     rax, rax
        jmp     @d
    @c:
        lea     r8, [r8 + 8]
        xor     rax, rax
        inc     rax
        mov     qword ptr [r8], rax
    @d:
end;
{$ELSE}
// eax | @a1    | ebx
// edx | @a2    |
// ecx | @a3    |
//     | XCount | esi
//     | DCount | esi
asm
        push    ebx
        push    esi
     // push    edi

        mov     ebx, eax
        mov     esi, [XCount]

        clc

        mov     eax, dword ptr [ebx]
     // mov     edi, dword ptr [edx]
     // add     eax, edi
        add     eax, dword ptr [edx]
        mov     dword ptr [ecx], eax

        dec     esi
        jz      @b1
    @a1:
        lea     edx, [edx + 4]
        lea     ebx, [ebx + 4]
        lea     ecx, [ecx + 4]
        mov     eax, dword ptr [ebx]
     // mov     edi, dword ptr [edx]
     // adc     eax, edi
        adc     eax, dword ptr [edx]
        mov     dword ptr [ecx], eax
        dec     esi
        jnz     @a1
    @b1:
        mov     esi, [DCount]
        dec     esi
        jl      @b2
        mov     edx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     edx, edx
        // *)
    @a2:
        lea     ebx, [ebx + 4]
        lea     ecx, [ecx + 4]
        mov     eax, dword ptr [ebx]
        adc     eax, edx
        mov     dword ptr [ecx], eax
        dec     esi
        jge     @a2
    @b2:
        jc      @c
        xor     eax, eax
        jmp     @d
    @c:
        lea     ecx, [ecx + 4]
        xor     eax, eax
        inc     eax
        mov     dword ptr [ecx], eax
    @d:
     // pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 + a2
  Count = a1.Count
}
function BNX_add_int_asm (a1: PUIntX; a2: UIntX; a3: PUIntX; Count: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1   |
// rdx | a2    |
// r8  | @a3   |
// r9  | Count |
asm
        clc

        mov     rax, qword ptr [rcx]
        add     rax, rdx
        mov     qword ptr [r8], rax

        dec     r9
        jz      @b
        mov     rdx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     rdx, rdx
        // *)
    @a:
        lea     rcx, [rcx + 8]
        lea     r8, [r8 + 8]
        mov     rax, qword ptr [rcx]
        adc     rax, rdx
        mov     qword ptr [r8], rax
        dec     r9
        jnz     @a
    @b:
        jc      @c
        xor     rax, rax
        jmp     @d
    @c:
        lea     r8, [r8 + 8]
        xor     rax, rax
        inc     rax
        mov     qword ptr [r8], rax
    @d:
end;
{$ELSE}
// eax | @a1   | ebx
// edx | a2    |
// ecx | @a3   |
//     | Count | esi
asm
        push    ebx
        push    esi

        mov     ebx, eax
        mov     esi, [Count]

        clc

        mov     eax, dword ptr [ebx]
        add     eax, edx
        mov     dword ptr [ecx], eax

        dec     esi
        jz      @b
        mov     edx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     edx, edx
        // *)
    @a:
        lea     ebx, [ebx + 4]
        lea     ecx, [ecx + 4]
        mov     eax, dword ptr [ebx]
        adc     eax, edx
        mov     dword ptr [ecx], eax
        dec     esi
        jnz     @a
    @b:
        jc      @c
        xor     eax, eax
        jmp     @d
    @c:
        lea     ecx, [ecx + 4]
        xor     eax, eax
        inc     eax
        mov     dword ptr [ecx], eax
    @d:
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 - a2
  |a1| > |a2|
  XCount = min (a1.Count, a2.Count) = a2.Count
  DCount = a1.Count - a2.Count
}
function BNX_sub_asm (a1, a2, a3: PUIntX; XCount, DCount: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1    |
// rdx | @a2    |
// r8  | @a3    |
// r9  | XCount |
//     | DCount | r9
asm
        clc

        mov     rax, qword ptr [rcx]
     // mov     r10, qword ptr [rdx]
     // sub     rax, r10
        sub     rax, qword ptr [rdx]
        mov     qword ptr [r8], rax

        dec     r9
        jz      @b1
    @a1:
        lea     rdx, [rdx + 8]
        lea     rcx, [rcx + 8]
        lea     r8, [r8 + 8]
        mov     rax, qword ptr [rcx]
     // mov     r10, qword ptr [rdx]
     // sbb     rax, r10
        sbb     rax, qword ptr [rdx]
        mov     qword ptr [r8], rax
        dec     r9
        jnz     @a1
    @b1:
        mov     r9, [DCount]
        dec     r9
        jl      @b2
        mov     rdx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     rdx, rdx
        // *)
    @a2:
        lea     rcx, [rcx + 8]
        lea     r8, [r8 + 8]
        mov     rax, qword ptr [rcx]
        sbb     rax, rdx
        mov     qword ptr [r8], rax
        dec     r9
        jge     @a2
    @b2:
end;
{$ELSE}
// eax | @a1    | ebx
// edx | @a2    |
// ecx | @a3    |
//     | XCount | esi
//     | DCount | esi
asm
        push    ebx
        push    esi
     // push    edi

        mov     ebx, eax
        mov     esi, [XCount]

        clc

        mov     eax, dword ptr [ebx]
     // mov     edi, dword ptr [edx]
     // sub     eax, edi
        sub     eax, dword ptr [edx]
        mov     dword ptr [ecx], eax

        dec     esi
        jz      @b1
    @a1:
        lea     edx, [edx + 4]
        lea     ebx, [ebx + 4]
        lea     ecx, [ecx + 4]
        mov     eax, dword ptr [ebx]
     // mov     edi, dword ptr [edx]
     // sbb     eax, edi
        sbb     eax, dword ptr [edx]
        mov     dword ptr [ecx], eax
        dec     esi
        jnz     @a1
    @b1:
        mov     esi, [DCount]
        dec     esi
        jl      @b2
        mov     edx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     edx, edx
        // *)
    @a2:
        lea     ebx, [ebx + 4]
        lea     ecx, [ecx + 4]
        mov     eax, dword ptr [ebx]
        sbb     eax, edx
        mov     dword ptr [ecx], eax
        dec     esi
        jge     @a2
    @b2:
     // pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 - a2
  |a1| > |a2|
  Count = a1.Count
}
function BNX_sub_int_asm (a1: PUIntX; a2: UIntX; a3: PUIntX; Count: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1   |
// rdx | a2    |
// r8  | @a3   |
// r9  | Count |
asm
        clc

        mov     rax, qword ptr [rcx]
        sub     rax, rdx
        mov     qword ptr [r8], rax

        dec     r9
        jz      @b
        mov     rdx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     rdx, rdx
        // *)
    @a:
        lea     rcx, [rcx + 8]
        lea     r8, [r8 + 8]
        mov     rax, qword ptr [rcx]
        sbb     rax, rdx
        mov     qword ptr [r8], rax
        dec     r9
        jnz     @a
    @b:
end;
{$ELSE}
// eax | @a1   | ebx
// edx | a2    |
// ecx | @a3   |
//     | Count | esi
asm
        push    ebx
        push    esi

        mov     ebx, eax
        mov     esi, [Count]

        clc

        mov     eax, dword ptr [ebx]
        sub     eax, edx
        mov     dword ptr [ecx], eax

        dec     esi
        jz      @b
        mov     edx, 0
        (* xor  ,     CF
           xor cannot be used because the CF flag is reset
        xor     edx, edx
        // *)
    @a:
        lea     ebx, [ebx + 4]
        lea     ecx, [ecx + 4]
        mov     eax, dword ptr [ebx]
        sbb     eax, edx
        mov     dword ptr [ecx], eax
        dec     esi
        jnz     @a
    @b:
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 + a2
}
procedure BN_add (var a3: T_BNX; const a1, a2: T_BNX);
var
  P1, P2   : PUIntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VCount3  : IntX;
  VCount   : IntX;
  VXCount  : IntX;
  VDCount  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_asm_proc;
  VValue   : UIntX;
begin
  VSign1 := a1.Sign;
  VSign2 := a2.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = a2
    BN_dup (a3, a2);
    Exit;
  end;

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // a2 = 0
    // a3 = a1
    BN_dup (a3, a1);
    Exit;
  end;

  VCount := VCount1;
  VXCount := VCount1;
  VDCount := 0;

  if VCount1 < VCount2 then begin
    VCount := VCount2;
    // VXCount := VCount1;
    VDCount := VCount2 - VCount1;
  end else
  if VCount2 < VCount1 then begin
    // VCount := VCount1;
    VXCount := VCount2;
    VDCount := VCount1 - VCount2;
  end;

  VCompare := BN_DataCompare (a1, a2);

  if VSign1 = VSign2 then begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| + |a2| = 2 * |a1|
      // ----
      // (+3) + (+3) = +6
      // (-3) + (-3) = -6
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      BN_dup (a3, a1); // a3 = a1
      BN_shl_(a3, 1);  // a3 = 2 * a1
      Exit;
    end;

    // |a3| = |a1| + |a2|
    // ----
    // (+3) + (+2) = +5
    // (-3) + (-2) = -5
    // (+2) + (+3) = +5
    // (-2) + (-3) = -5
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount + 1;

    if VCompare > 0 then begin
      // |a1| > |a2|
      P1 := a1.PLData;
      P2 := a2.PLData;
    end else begin
      // |a1| < |a2|
      P1 := a2.PLData;
      P2 := a1.PLData;
    end;

    VProc := BNX_add_asm;
  end else begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    VCount3 := VCount;

    if VCompare > 0 then begin
      // |a1| > |a2|
      // |a3| = |a1| - |a2|
      // ----
      // (+3) + (-2) = +1
      // (-3) + (+2) = -1
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      P1 := a1.PLData;
      P2 := a2.PLData;
    end else begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // ----
      // (+2) + (-3) = -1
      // (-2) + (+3) = +1
      // a3.sign = a2.sign

      VSign3 := VSign2;

      P1 := a2.PLData;
      P2 := a1.PLData;
    end;

    VProc := BNX_sub_asm;
  end;

  BN_Init (a3, VSign3, VCount3);
  VValue := VProc (P1, P2, a3.PLData, VXCount, VDCount);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1 = a1 + a2
}
procedure BN_add_(var a1: T_BNX; const a2: T_BNX);
var
  V : T_BNX;
begin
  BN_dup (V, a1);
  BN_add (a1, V, a2);
  BN_Clear (V);
end;

{
  a3 = a1 + a2
}
procedure BN_add_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  VSign1   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount3  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_int_asm_proc;
  VValue   : UIntX;
begin
  VSign1 := a1.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = a2
    BN_AssignInt (a3, ASign2, a2);
    Exit;
  end;

  if (ASign2 = 0) or (a2 = 0) then begin
    // a2 = 0
    // a3 = a1
    BN_dup (a3, a1);
    Exit;
  end;

  if VSign1 = ASign2 then begin
    // |a3| = |a1| + |a2|
    // ----
    // (+3) + (+2) = +5
    // (-3) + (-2) = -5
    // (+2) + (+3) = +5
    // (-2) + (-3) = -5
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1 + 1;

    VProc := BNX_add_int_asm;
  end else begin
    VCompare := BN_DataCompare (a1, a2);

    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    if VCompare < 0 then begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // VCount1 = 1
      // ----
      // (+2) + (-3) = -1
      // (-2) + (+3) = +1
      // a3.sign = a2.sign

      // VSign3 := ASign2;

      BN_AssignInt (a3, ASign2, a2 - a1.Data[0]);
      Exit;
    end;

    // |a1| > |a2|
    // |a3| = |a1| - |a2|
    // VCount1 > 1
    // ----
    // (+3) + (-2) = +1
    // (-3) + (+2) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1;

    VProc := BNX_sub_int_asm;
  end;

  BN_Init (a3, VSign3, VCount3);
  VValue := VProc (a1.PLData, a2, a3.PLData, VCount1);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1 = a1 + a2
}
procedure BN_add_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  V : T_BNX;
begin
  BN_dup (V, a1);
  BN_add_int (a1, V, ASign2, a2);
  BN_Clear (V);
end;

{
  a3 = a1 + a2
  a2 - signed integer
}
procedure BN_add_sint (var a3: T_BNX; const a1: T_BNX; a2: IntX);
var
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_add_int (a3, a1, VSign2, a2);
end;

{
  a1 = a1 + a2
  a2 - signed integer
}
procedure BN_add_sint_(var a1: T_BNX; a2: IntX);
var
  V      : T_BNX;
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_dup (V, a1);
  BN_add_int (a1, V, VSign2, a2);
  BN_Clear (V);
end;

{
  a3 = a1 - a2
}
procedure BN_sub (var a3: T_BNX; const a1, a2: T_BNX);
var
  P1, P2   : PUIntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VCount3  : IntX;
  VCount   : IntX;
  VXCount  : IntX;
  VDCount  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_asm_proc;
  VValue   : UIntX;
begin
  VSign1 := a1.Sign;
  VSign2 := a2.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = -a2
    BN_ndup (a3, a2);
    Exit;
  end;

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // a2 = 0
    // a3 = a1
    BN_dup (a3, a1);
    Exit;
  end;

  VCount := VCount1;
  VXCount := VCount1;
  VDCount := 0;

  if VCount1 < VCount2 then begin
    VCount := VCount2;
    // VXCount := VCount1;
    VDCount := VCount2 - VCount1;
  end else
  if VCount2 < VCount1 then begin
    // VCount := VCount1;
    VXCount := VCount2;
    VDCount := VCount1 - VCount2;
  end;

  VCompare := BN_DataCompare (a1, a2);

  if VSign1 = VSign2 then begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    VCount3 := VCount;

    if VCompare > 0 then begin
      // |a1| > |a2|
      // |a3| = |a1| - |a2|
      // ----
      // (+3) - (+2) = +1
      // (-3) - (-2) = -1
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      P1 := a1.PLData;
      P2 := a2.PLData;
    end else begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // ----
      // (+2) - (+3) = -1
      // (-2) - (-3) = +1
      // a3.sign = -a1.sign

      VSign3 := -VSign1;

      P1 := a2.PLData;
      P2 := a1.PLData;
    end;

    VProc := BNX_sub_asm;
  end else begin
    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| + |a2| = 2 * |a1|
      // ----
      // (+3) - (-3) = +6
      // (-3) - (+3) = -6
      // a3.sign = a1.sign

      // VSign3 := VSign1;

      BN_dup (a3, a1); // a3 = a1
      BN_shl_(a3, 1);  // a3 = 2 * a1
      Exit;
    end;

    // |a3| = |a1| + |a2|
    // ----
    // (+3) - (-2) = +5
    // (-3) - (+2) = -1
    // (+2) - (-3) = +5
    // (-2) - (+3) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount + 1;

    if VCompare > 0 then begin
      // |a1| > |a2|
      P1 := a1.PLData;
      P2 := a2.PLData;
    end else begin
      // |a1| < |a2|
      P1 := a2.PLData;
      P2 := a1.PLData;
    end;

    VProc := BNX_add_asm;
  end;

  BN_Init (a3, VSign3, VCount3);
  VValue := VProc (P1, P2, a3.PLData, VXCount, VDCount);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1 = a1 - a2
}
procedure BN_sub_(var a1: T_BNX; const a2: T_BNX);
var
  V : T_BNX;
begin
  BN_dup (V, a1);
  BN_sub (a1, V, a2);
  BN_Clear (V);
end;

{
  a3 = a1 - a2
}
procedure BN_sub_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  VSign1   : ShortInt;
  VSign3   : ShortInt;
  VCount1  : IntX;
  VCount3  : IntX;
  VCompare : Integer;
  VProc    : T_BNX_add_sub_int_asm_proc;
  VValue   : UIntX;
begin
  VSign1 := a1.Sign;
  VSign3 := VSign1;

  VCount1 := BN_DataCount (a1);

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // a1 = 0
    // a3 = -a2
    BN_AssignInt (a3, -ASign2, a2);
    Exit;
  end;

  if (ASign2 = 0) or (a2 = 0) then begin
    // a2 = 0
    // a3 = a1
    BN_dup (a3, a1);
    Exit;
  end;

  if VSign1 = ASign2 then begin
    VCompare := BN_DataCompare (a1, a2);

    if VCompare = 0 then begin
      // |a1| = |a2|
      // |a3| = |a1| - |a2| = 0
      BN_AssignZero (a3);
      Exit;
    end;

    if VCompare < 0 then begin
      // |a1| < |a2|
      // |a3| = |a2| - |a1|
      // VCount1 = 1
      // ----
      // (-2) - (-3) = +1
      // (+2) - (+3) = -1
      // a3.sign = -a1.sign

      // VSign3 := -VSign1;

      BN_AssignInt (a3, -VSign1, a2 - a1.Data[0]);
      Exit;
    end;

    // |a1| > |a2|
    // |a3| = |a1| - |a2|
    // VCount1 > 1
    // ----
    // (+3) - (+2) = +1
    // (-3) - (-2) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1;

    VProc := BNX_sub_int_asm;
  end else begin
    // |a3| = |a1| + |a2|
    // ----
    // (+3) - (-2) = +5
    // (-3) - (+2) = -1
    // (+2) - (-3) = +5
    // (-2) - (+3) = -1
    // a3.sign = a1.sign

    // VSign3 := VSign1;
    VCount3 := VCount1 + 1;

    VProc := BNX_add_int_asm;
  end;

  BN_Init (a3, VSign3, VCount3);
  VValue := VProc (a1.PLData, a2, a3.PLData, VCount1);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1 = a1 - a2
}
procedure BN_sub_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  V : T_BNX;
begin
  BN_dup (V, a1);
  BN_sub_int (a1, V, ASign2, a2);
  BN_Clear (V);
end;

{
  a3 = a1 - a2
  a2 - signed integer
}
procedure BN_sub_sint (var a3: T_BNX; const a1: T_BNX; a2: IntX);
var
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_sub_int (a3, a1, VSign2, a2);
end;

{
  a1 = a1 - a2
  a2 - signed integer
}
procedure BN_sub_sint_(var a1: T_BNX; a2: IntX);
var
  V      : T_BNX;
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_dup (V, a1);
  BN_sub_int (a1, V, VSign2, a2);
  BN_Clear (V);
end;

{
     ()     ASign1  ASign2
  Return the sign of the product (division) of two numbers with signs ASign1 and ASign2
}
function BN_mul_div_sign (ASign1, ASign2: ShortInt): ShortInt;
{$IFDEF CPUX64}
// rcx | ASign1
// rdx | ASign2
asm
        xor     rax, rax

        test    cl, cl
        jz      @r       // ASign1 = 0
        test    dl, dl
        jz      @r       // ASign2 = 0
        xor     cl, dl
        jns     @b       // ASign1 = ASign2

 // @a: dec     rax      // ASign1 <> ASign2
 //     jmp     @r
    @a: mov     rax, -2  // ASign1 <> ASign2

    @b: inc     rax      // ASign1 = ASign2
    @r:
end;
{$ELSE}
// eax | ASign1 | ecx
// edx | ASign2
asm
        mov     ecx, eax
        xor     eax, eax

        test    cl, cl
        jz      @r       // ASign1 = 0
        test    dl, dl
        jz      @r       // ASign2 = 0
        xor     cl, dl
        jns     @b       // ASign1 = ASign2

 // @a: dec     eax      // ASign1 <> ASign2
 //     jmp     @r
    @a: mov     eax, -2  // ASign1 <> ASign2

    @b: inc     eax      // ASign1 = ASign2
    @r:
end;
{$ENDIF}

{
     ()     ASign1  ASign2
  Return the sign of the product (division) of two numbers with signs ASign1 and ASign2
}
function BN_mul_div_sign_(ASign1: ShortInt; ASign2: ShortInt): ShortInt;
begin
  Result := 0;

  if (ASign1 = 0) or (ASign2 = 0) then Exit;

  if ASign1 = ASign2 then begin
    Inc (Result);
    Exit;
  end;

  // ASign1 <> ASign2
  Dec (Result);

  (*
  if (ASign1 = 0) or (ASign2 = 0) then begin
    Result := 0;
  end else
  if (ASign1 = ASign2) then begin
    Result := 1;
  end else begin
    Result := -1;
  end;
  // *)

  (*
  Result := ASign1 * ASign2;
  // *)
end;

{
  a3 = a1 * a2
}
procedure BNX_mul_int_asm (a1, a2: UIntX; var a3: UIntX; var Carry: UIntX);
{$IFDEF CPUX64}
// rcx | a1
// rdx | a2
// r8  | @a3
// r9  | @Carry
asm
        mov     rax, rcx
        mul     rdx

        add     rax, qword ptr [r9]
        jnc     @a
        inc     rdx
    @a:
        add     rax, qword ptr [r8]
        jnc     @b
        inc     rdx
    @b:
        mov     qword ptr [r8], rax
        mov     qword ptr [r9], rdx
end;
{$ELSE}
// eax | a1     |
// edx | a2     |
// ecx | @a3    |
//     | @Carry | ebx
asm
        push    ebx

        mov     ebx, [Carry]
        mul     edx

        add     eax, dword ptr [ebx]
        jnc     @a
        inc     edx
    @a:
        add     eax, dword ptr [ecx]
        jnc     @b
        inc     edx
    @b:
        mov     dword ptr [ecx], eax
        mov     dword ptr [ebx], edx

        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 * a2
  Count1 = a1.Count
  Count2 = a2.Count
}
function BNX_mul_base_asm (a1, a2, a3: PUIntX; Count1, Count2: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1    | r10
// rdx | @a2    | r11
// r8  | @a3    | r12
// r9  | Count1 | r13
//     | Count2 | r14
//     | Carry  | r15
asm
     // push    rbx
        push    r12
        push    r13
        push    r14
        push    r15

        mov     r10, rcx
        mov     r11, rdx
        mov     r12, r8
        mov     r13, r9
        mov     r14, [Count2]

    @loop_I:
        xor     r15, r15 // Carry = 0
    @loop_J:
        mov     rax, qword ptr [rcx]
     // mov     rbx, qword ptr [r11]
     // mul     rbx // rax * a2
        mul     qword ptr [r11] // rax * a2

        add     rax, r15
        jnc     @a
        inc     rdx
    @a:
     // mov     rbx, qword ptr [r8]
     // add     rax, rbx
        add     rax, qword ptr [r8]
        jnc     @b
        inc     rdx
    @b:
        mov     qword ptr [r8], rax
        mov     r15, rdx
    @next_J:
        lea     rcx, [rcx + 8] // Inc (@a1)
        lea     r8, [r8 + 8]   // Inc (@a3)
        dec     r9
        jnz     @loop_J
    @out_J:

        mov     rax, qword ptr [r8]
        add     rax, r15
        mov     qword ptr [r8], rax

    @next_I:
        mov     rcx, r10       // Restore @a1
        lea     r11, [r11 + 8] // Inc (@a2)
        lea     r12, [r12 + 8] // Inc (@a3)
        mov     r8, r12
        mov     r9, r13        // Restore Loop_J counter
        dec     r14
        jnz     @loop_I
        (*
        dec     r14
        jz      @out_I
        mov     rcx, r10       // Restore @a1
        lea     r11, [r11 + 8] // Inc (@a2)
        lea     r12, [r12 + 8] // Inc (@a3)
        mov     r8, r12
        mov     r9, r13        // Restore Loop_J counter
        jmp     @loop_I
        // *)
    @out_I:

        pop     r15
        pop     r14
        pop     r13
        pop     r12
     // pop     rbx
end;
{$ELSE}
// eax | @a1    | ebp [esp + 04]
// edx | @a2    | esi
// ecx | @a3    | edi
//     | Count1 | ebx [esp + 08]
//     | Count2 |     [esp + 12]
//     | Carry  |     [esp]
asm
        push    ebx
        push    esi
        push    edi
        push    ebp

        sub     esp, 16 // Alloc Local variables

        mov     esi, edx
        mov     edi, ecx

        mov     ebx, [Count2]
        mov     dword ptr [esp + 12], ebx // I
        mov     ebx, [Count1]
        mov     dword ptr [esp + 08], ebx // J

        mov     ebp, eax
        mov     dword ptr [esp + 04], eax

    @loop_I:
        mov     dword ptr [esp], 0 // Carry = 0
    @loop_J:
        mov     eax, dword ptr [ebp]
        mul     dword ptr [esi] // eax * a2

        add     eax, dword ptr [esp]
        jnc     @a
        inc     edx
    @a:
        add     eax, dword ptr [ecx]
        jnc     @b
        inc     edx
    @b:
        mov     dword ptr [ecx], eax
        mov     dword ptr [esp], edx
    @next_J:
        lea     ebp, [ebp + 4] // Inc (@a1)
        lea     ecx, [ecx + 4] // Inc (@a3)
        dec     ebx
        jnz     @loop_J
    @out_J:

        mov     eax, dword ptr [ecx]
        add     eax, dword ptr [esp]
        mov     dword ptr [ecx], eax

    @next_I:
        mov     ebp, dword ptr [esp + 04] // Restore @a1
        lea     esi, [esi + 4]            // Inc (@a2)
        lea     edi, [edi + 4]            // Inc (@a3)
        mov     ecx, edi
        mov     ebx, dword ptr [esp + 08] // Restore Loop_J counter
        dec     dword ptr [esp + 12]
        jnz     @loop_I
        (*
        dec     dword ptr [esp + 12]
        jz      @out_I
        mov     ebp, dword ptr [esp + 04] // Restore @a1
        lea     esi, [esi + 4]            // Inc (@a2)
        lea     edi, [edi + 4]            // Inc (@a3)
        mov     ecx, edi
        mov     ebx, dword ptr [esp + 08] // Restore Loop_J counter
        jmp     @loop_I
        // *)
    @out_I:

        add     esp, 16 // Free Local variables

        pop     ebp
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 * a2
  Count1 = a1.Count
}
function BNX_mul_int_base_asm (a1: PUIntX; a2: UIntX; a3: PUIntX; Count1: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1    |
// rdx | a2     | r10
// r8  | @a3    |
// r9  | Count1 |
//     | Carry  | r11
asm
        mov     r10, rdx
        xor     r11, r11

    @loop_J:
        mov     rax, qword ptr [rcx]
        mul     r10

        add     rax, r11
        jnc     @a
        inc     rdx
    @a:
     // mov     r12, qword ptr [r8]
     // add     rax, r12
        add     rax, qword ptr [r8]
        jnc     @b
        inc     rdx
    @b:
        mov     qword ptr [r8], rax
        mov     r11, rdx
    @next_J:
        lea     rcx, [rcx + 8]
        lea     r8, [r8 + 8]
        dec     r9
        jnz     @loop_J
    @out_J:

        mov     rax, qword ptr [r8]
        add     rax, r11
        mov     qword ptr [r8], rax
end;
{$ELSE}
// eax | @a1    | ebp
// edx | a2     | edi
// ecx | @a3    |
//     | Count1 | ebx
//     | Carry  | esi
asm
        push    ebx
        push    esi
        push    edi
        push    ebp

        mov     ebx, [Count1]

        mov     ebp, eax
        mov     edi, edx
        xor     esi, esi

    @loop_J:
        mov     eax, dword ptr [ebp]
        mul     edi

        add     eax, esi
        jnc     @a
        inc     edx
    @a:
        add     eax, dword ptr [ecx]
        jnc     @b
        inc     edx
    @b:
        mov     dword ptr [ecx], eax
        mov     esi, edx
    @next_J:
        lea     ebp, [ebp + 4]
        lea     ecx, [ecx + 4]
        dec     ebx
        jnz     @loop_J
    @out_J:

        mov     eax, dword ptr [ecx]
        add     eax, esi
        mov     dword ptr [ecx], eax

        pop     ebp
        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  a3 = a1 * a2
}
procedure BN_mul_base (var a3: T_BNX; const a1, a2: T_BNX);
var
  VUInt   : UIntX;
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount2 : IntX;
  VCount3 : IntX;
  VValue  : UIntX;
begin
  VSign3 := BN_mul_div_sign (a1.Sign, a2.Sign);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VCount1 <= 0) or (VCount2 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  (*
  BN_mul_base_nz (a3, a1, a2, VCount1, VCount2, VSign3);
  Exit;
  // *)

  if BN_IsInt (a1, VUInt) then begin
    BN_mul_int_nz (a3, a2, VUInt, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a2, VUInt) then begin
    BN_mul_int_nz (a3, a1, VUInt, VCount1, VSign3);
    Exit;
  end;

  VCount3 := VCount1 + VCount2;
  BN_Init (a3, VSign3, VCount3);

  VValue := BNX_mul_base_asm (a1.PLData, a2.PLData, a3.PLData, VCount1, VCount2);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1 = a1 * a2
}
procedure BN_mul_base_(var a1: T_BNX; const a2: T_BNX);
var
  V : T_BNX;
begin
  BN_mul_base (V, a1, a2);
  BN_dup (a1, V);
  BN_Clear (V);
end;

{
  a3 = a1 * a2
  a1 <> 0
  a2 <> 0
  a1.Count = ACount1
  a2.Count = ACount2
  a3.Sign  = ASign3
}
procedure BN_mul_base_nz (var a3: T_BNX; const a1, a2: T_BNX;
  ACount1, ACount2: IntX; ASign3: ShortInt);
var
  VUInt   : UIntX;
  VCount3 : IntX;
  VValue  : UIntX;
begin
  if BN_IsInt (a1, VUInt) then begin
    BN_mul_int_nz (a3, a2, VUInt, ACount2, ASign3);
    Exit;
  end;

  if BN_IsInt (a2, VUInt) then begin
    BN_mul_int_nz (a3, a1, VUInt, ACount1, ASign3);
    Exit;
  end;

  VCount3 := ACount1 + ACount2;
  BN_Init (a3, ASign3, VCount3);

  VValue := BNX_mul_base_asm (a1.PLData, a2.PLData, a3.PLData, ACount1, ACount2);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1(a3) = a1 * a2
  a1 <> 0
  a2 <> 0
  a1.Count = ACount1
  a2.Count = ACount2
  a3.Sign  = ASign3
}
procedure BN_mul_base_nz_(var a1: T_BNX; const a2: T_BNX;
  ACount1, ACount2: IntX; ASign3: ShortInt);
var
  V : T_BNX;
begin
  BN_mul_base_nz (V, a1, a2, ACount1, ACount2, ASign3);
  BN_dup (a1, V);
  BN_Clear (V);
end;

{
  a3 = a1 * a2
    
  Using the Karatsuba algorithm
}
procedure BN_mul (var a3: T_BNX; const a1, a2: T_BNX);
var
  VUInt   : UIntX;
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount2 : IntX;
  VmCount : IntX;
  Vm      : IntX;

  V_a0    : T_BNX;
  V_a1    : T_BNX;
  V_aa    : T_BNX;
  V_b0    : T_BNX;
  V_b1    : T_BNX;
  V_bb    : T_BNX;
  V_ab00  : T_BNX;
  V_ab11  : T_BNX;
  V_aabb  : T_BNX;
begin
  VSign3 := BN_mul_div_sign (a1.Sign, a2.Sign);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VCount1 <= 0) or (VCount2 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  {$IFDEF KARATSUBA_V2}
  // VmCount = min (VCount1, VCount2)
  if VCount1 >= VCount2 then VmCount := VCount2 else VmCount := VCount1;
  {$ELSE}
  // VmCount = max (VCount1, VCount2)
  if VCount1 >= VCount2 then VmCount := VCount1 else VmCount := VCount2;
  {$ENDIF}

  if VmCount < Karatsuba_mul_TCount then begin
    BN_mul_base_nz (a3, a1, a2, VCount1, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a1, VUInt) then begin
    BN_mul_int_nz (a3, a2, VUInt, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a2, VUInt) then begin
    BN_mul_int_nz (a3, a1, VUInt, VCount1, VSign3);
    Exit;
  end;

  {$IFDEF KARATSUBA_V2}
  if VCount1 = VCount2 then begin
    Vm := VmCount shr 1; // VmCount div 2;
  end else begin
    Vm := VmCount;
  end;
  {$ELSE}
  Vm := VmCount shr 1; // VmCount div 2;
  {$ENDIF}

  // (a0 + a1 * 2^[m]) * (b0 + b1 * 2^[m]) =
  // = (a0 * b0) + (a0 * b1 + a1 * b0) * 2^[m] + (a1 * b1) * 2^[2m]
  // | ab00 = (a0 * b0)
  // | ab11 = (a1 * b1)
  // | aabb = (a0 + a1) * (b0 + b1) = (a0 * b0) + (a1 * b1) + (a0 * b1 + a1 * b0)
  // | (a0 * b1 + a1 * b0) = (a0 + a1) * (b0 + b1) - ((a0 * b0) + (a1 * b1)) =
  // | = aabb - (ab00 + ab11)
  // = ab00 + (aabb - (ab00 + ab11)) *  2^[m] + ab11 * 2^[2m]

  BN_cdup (V_a0, a1, Vm);
  BN_hdup (V_a1, a1, Vm);
  BN_cdup (V_b0, a2, Vm);
  BN_hdup (V_b1, a2, Vm);

  BN_add (V_aa, V_a0, V_a1);     // V_aa = V_a0 + V_a1
  BN_add (V_bb, V_b0, V_b1);     // V_bb = V_b0 + V_b1

  BN_mul (V_ab00, V_a0, V_b0);   // V_ab00 = V_a0 * V_b0
  BN_mul (V_ab11, V_a1, V_b1);   // V_ab11 = V_a1 * V_b1
  BN_mul (V_aabb, V_aa, V_bb);   // V_aabb = (V_a0 + V_a1) * (V_b0 + V_b1)

  BN_add (V_aa, V_ab00, V_ab11); // V_aa = V_ab00 + V_ab11
  BN_sub (V_bb, V_aabb, V_aa);   // V_bb = V_aabb - (V_ab00 + V_ab11)
  BN_cshl_(V_bb, Vm);            // V_bb = (V_aabb - (V_ab00 + V_ab11)) * 2^[m]
  BN_cshl_(V_ab11, Vm shl 1);    // V_ab11 = V_ab11 * 2^[2m]
  BN_add (V_aa, V_ab00, V_bb);   // V_aa = V_ab00 + (V_aabb - (V_ab00 + V_ab11)) * 2^[m]
  BN_add (a3, V_aa, V_ab11);     // a3 = V_ab00 + (V_aabb - (V_ab00 + V_ab11)) * 2^[m] + V_ab11 * 2^[2m]
  a3.Sign := VSign3;

  BN_Clear (V_a0);
  BN_Clear (V_a1);
  BN_Clear (V_aa);
  BN_Clear (V_b0);
  BN_Clear (V_b1);
  BN_Clear (V_bb);
  BN_Clear (V_ab00);
  BN_Clear (V_ab11);
  BN_Clear (V_aabb);
end;

{
  a1 = a1 * a2
    
  Using the Karatsuba algorithm
}
procedure BN_mul_(var a1: T_BNX; const a2: T_BNX);
var
  VUInt   : UIntX;
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount2 : IntX;
  VmCount : IntX;
  Vm      : IntX;

  V_a0    : T_BNX;
  V_a1    : T_BNX;
  V_aa    : T_BNX;
  V_b0    : T_BNX;
  V_b1    : T_BNX;
  V_bb    : T_BNX;
  V_ab00  : T_BNX;
  V_ab11  : T_BNX;
  V_aabb  : T_BNX;
begin
  VSign3 := BN_mul_div_sign (a1.Sign, a2.Sign);

  if (VSign3 = 0) then begin
    BN_AssignZero (a1);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VCount1 <= 0) or (VCount2 <= 0) then begin
    BN_AssignZero (a1);
    Exit;
  end;

  {$IFDEF KARATSUBA_V2}
  // VmCount = min (VCount1, VCount2)
  if VCount1 >= VCount2 then VmCount := VCount2 else VmCount := VCount1;
  {$ELSE}
  // VmCount = max (VCount1, VCount2)
  if VCount1 >= VCount2 then VmCount := VCount1 else VmCount := VCount2;
  {$ENDIF}

  if VmCount < Karatsuba_mul_TCount then begin
    BN_mul_base_nz_(a1, a2, VCount1, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a1, VUInt) then begin
    BN_mul_int_nz (a1, a2, VUInt, VCount2, VSign3);
    Exit;
  end;

  if BN_IsInt (a2, VUInt) then begin
    BN_mul_int_nz_(a1, VUInt, VCount1, VSign3);
    Exit;
  end;

  {$IFDEF KARATSUBA_V2}
  if VCount1 = VCount2 then begin
    Vm := VmCount shr 1; // VmCount div 2;
  end else begin
    Vm := VmCount;
  end;
  {$ELSE}
  Vm := VmCount shr 1; // VmCount div 2;
  {$ENDIF}

  // (a0 + a1 * 2^[m]) * (b0 + b1 * 2^[m]) =
  // = (a0 * b0) + (a0 * b1 + a1 * b0) * 2^[m] + (a1 * b1) * 2^[2m]
  // | ab00 = (a0 * b0)
  // | ab11 = (a1 * b1)
  // | aabb = (a0 + a1) * (b0 + b1) = (a0 * b0) + (a1 * b1) + (a0 * b1 + a1 * b0)
  // | (a0 * b1 + a1 * b0) = (a0 + a1) * (b0 + b1) - ((a0 * b0) + (a1 * b1)) =
  // | = aabb - (ab00 + ab11)
  // = ab00 + (aabb - (ab00 + ab11)) *  2^[m] + ab11 * 2^[2m]

  BN_cdup (V_a0, a1, Vm);
  BN_hdup (V_a1, a1, Vm);
  BN_cdup (V_b0, a2, Vm);
  BN_hdup (V_b1, a2, Vm);

  BN_add (V_aa, V_a0, V_a1);     // V_aa = V_a0 + V_a1
  BN_add (V_bb, V_b0, V_b1);     // V_bb = V_b0 + V_b1

  BN_mul (V_ab00, V_a0, V_b0);   // V_ab00 = V_a0 * V_b0
  BN_mul (V_ab11, V_a1, V_b1);   // V_ab11 = V_a1 * V_b1
  BN_mul (V_aabb, V_aa, V_bb);   // V_aabb = (V_a0 + V_a1) * (V_b0 + V_b1)

  BN_add (V_aa, V_ab00, V_ab11); // V_aa = V_ab00 + V_ab11
  BN_sub (V_bb, V_aabb, V_aa);   // V_bb = V_aabb - (V_ab00 + V_ab11)
  BN_cshl_(V_bb, Vm);            // V_bb = (V_aabb - (V_ab00 + V_ab11)) * 2^[m]
  BN_cshl_(V_ab11, Vm shl 1);    // V_ab11 = V_ab11 * 2^[2m]
  BN_add (V_aa, V_ab00, V_bb);   // V_aa = V_ab00 + (V_aabb - (V_ab00 + V_ab11)) * 2^[m]
  BN_add (a1, V_aa, V_ab11);     // a1 = V_ab00 + (V_aabb - (V_ab00 + V_ab11)) * 2^[m] + V_ab11 * 2^[2m]
  a1.Sign := VSign3;

  BN_Clear (V_a0);
  BN_Clear (V_a1);
  BN_Clear (V_aa);
  BN_Clear (V_b0);
  BN_Clear (V_b1);
  BN_Clear (V_bb);
  BN_Clear (V_ab00);
  BN_Clear (V_ab11);
  BN_Clear (V_aabb);
end;

{
  a3 = a1 * a2
}
procedure BN_mul_int (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount3 : IntX;
  VValue  : UIntX;
begin
  if (a2 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VSign3 := BN_mul_div_sign (a1.Sign, ASign2);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);

  if (VCount1 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount3 := VCount1 + 1;
  BN_Init (a3, VSign3, VCount3);

  VValue := BNX_mul_int_base_asm (a1.PLData, a2, a3.PLData, VCount1);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1 = a1 * a2
}
procedure BN_mul_int_(var a1: T_BNX; ASign2: ShortInt; a2: UIntX);
var
  V : T_BNX;
begin
  BN_mul_int (V, a1, ASign2, a2);
  BN_dup (a1, V);
  BN_Clear (V);
end;

{
  a3 = a1 * a2
  a1 <> 0
  a2 <> 0
  a1.Count = ACount1
  a3.Sign  = ASign3
}
procedure BN_mul_int_nz (var a3: T_BNX; const a1: T_BNX; a2: UIntX;
  ACount1: IntX; ASign3: ShortInt);
var
  VCount3 : IntX;
  VValue  : UIntX;
begin
  VCount3 := ACount1 + 1;
  BN_Init (a3, ASign3, VCount3);

  VValue := BNX_mul_int_base_asm (a1.PLData, a2, a3.PLData, ACount1);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a1(a3) = a1 * a2
  a1 <> 0
  a2 <> 0
  a1.Count = ACount1
  a3.Sign  = ASign3
}
procedure BN_mul_int_nz_(var a1: T_BNX; a2: UIntX;
  ACount1: IntX; ASign3: ShortInt);
var
  V : T_BNX;
begin
  BN_mul_int_nz (V, a1, a2, ACount1, ASign3);
  BN_dup (a1, V);
  BN_Clear (V);
end;

{
  a3 = a1 * a2
  a2 - signed integer
}
procedure BN_mul_sint (var a3: T_BNX; const a1: T_BNX; a2: IntX);
var
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_mul_int (a3, a1, VSign2, a2);
end;

{
  a3 = a1 * a2'
  a2'[n] = a2,   (other elements are zeros)
}
procedure BN_mul_intn (var a3: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX;
  n: IntX);
var
  VSign3  : ShortInt;
  VCount1 : IntX;
  VCount3 : IntX;
  VValue  : UIntX;
begin
  if (a2 = 0) or (n < 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VSign3 := BN_mul_div_sign (a1.Sign, ASign2);

  if (VSign3 = 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount1 := BN_DataCount (a1);

  if (VCount1 <= 0) then begin
    BN_AssignZero (a3);
    Exit;
  end;

  VCount3 := VCount1 + n + 1;
  BN_Init (a3, VSign3, VCount3);

  VValue := BNX_mul_int_base_asm (a1.PLData, a2, @a3.Data[n], VCount1);
  if VValue = 0 then BN_Pack (a3);
end;

{
  a3 = a1 * a2'
  a2'[n] = a2,   (other elements are zeros)
  a2 - signed integer
}
procedure BN_mul_sintn (var a3: T_BNX; const a1: T_BNX; a2: IntX; n: IntX);
var
  VSign2 : ShortInt;
begin
  VSign2 := 0;
  if a2 > 0 then VSign2 := 1 else
  if a2 < 0 then begin
    VSign2 := -1;
    a2 := -a2;
  end;

  BN_mul_intn (a3, a1, VSign2, a2, n);
end;

{
}
procedure BN8_mul_asm (a1, a2: UInt8; var a3: UInt8; var Carry: UInt8);
{$IFDEF CPUX64}
// rcx | a1
// rdx | a2
// r8  | @a3
// r9  | @Carry
asm
        mov     al, cl
        mul     dl
        add     al, byte ptr [r9]
        jnc     @a
        {$IFDEF X64_INC_AH}
        inc     ah
        {$ELSE}
        db      $FE
        db      $C4
        {$ENDIF}
    @a:
        add     al, byte ptr [r8]
        jnc     @b
        {$IFDEF X64_INC_AH}
        inc     ah
        {$ELSE}
        db      $FE
        db      $C4
        {$ENDIF}
    @b:
        mov     byte ptr [r8], al
     // mov     byte ptr [r9], ah // ERROR
        shr     ax, 8
        mov     byte ptr [r9], al
    @c:
end;
{$ELSE}
// eax | a1     |
// edx | a2     |
// ecx | @a3    |
//     | @Carry | ebx
asm
        push    ebx

        mov     ebx, [Carry]

        mul     dl
        add     al, byte ptr [ebx]
        jnc     @a
        inc     ah
    @a:
        add     al, byte ptr [ecx]
        jnc     @b
        inc     ah
    @b:
        mov     byte ptr [ecx], al
        mov     byte ptr [ebx], ah
    @c:

        pop     ebx
end;
{$ENDIF}

{
  c = a div b
  Result = c
}
function BNX_div_asm (aL, aH, b: UIntX; c: PUIntX): UIntX;
{$IFDEF CPUX64}
// rcx | aL    | rax  | adiv
// rdx | aH    | amod
// r8  | b     |
// r9  | @c    |
asm
        mov     rax, rcx
        div     r8
        mov     [r9], rax
end;
{$ELSE}
// eax | aL    | adiv
// edx | aH    | amod
// ecx | b     |
//     | @c    | ecx
asm
        div     ecx
        mov     ecx, [c]
        mov     [ecx], eax
end;
{$ENDIF}

{
  adiv = a div VDivisor
}
function BNX_div_intn (var adiv: T_BNX; const a: T_BNX;
  ADivisor: UIntX; ADIndex: IntX): UIntX;
{$IFDEF BN_INLINE} inline; {$ENDIF}
var
  VCount  : IntX;
  VRCount : IntX;
  VHIndex : IntX;
  VHiData : UIntX;
  VLoData : UIntX;
begin
  Result := 0;

  VCount := BN_DataCount (a);
  VHIndex := VCount - 1;
  VHiData := a.Data[VHIndex];

  (*
  if ADivisor = 1 then begin
    VRCount := VCount - ADIndex;
    VCount := BN_Init (adiv, a.Sign, VRCount);
    if VCount > 0 then begin
      Result := VHiData;
      adiv.PHData^ := VHiData;
    end;
    Exit;
  end else
  // *)

  if VHiData = ADivisor then begin
    VRCount := VCount - ADIndex;
    VCount := BN_Init (adiv, a.Sign, VRCount);
    if VCount > 0 then begin
      Result := 1;
      adiv.PHData^ := 1;
    end;
    Exit;
  end else

  if VHiData > ADivisor then begin
    VRCount := VCount - ADIndex;
    VCount := BN_Init (adiv, a.Sign, VRCount);
    if VCount > 0 then begin
      Result := BNX_div_asm (VHiData, 0, ADivisor, adiv.PHData);
    end;
    Exit;
  end;

  // VHiData < ADivisor
  VRCount := VHIndex - ADIndex; // VCount - ADIndex - 1;
  VCount := BN_Init (adiv, a.Sign, VRCount);
  if VCount > 0 then begin
    VLoData := a.Data[VHIndex - 1];
    Result := BNX_div_asm (VLoData, VHiData, ADivisor, adiv.PHData);
  end;
end;

{
  adiv = a1 div a2
  amod = a1 mod a2, amod > 0

  Result -  
  Result - number of iterations
}
function BN_div_mod (var adiv, amod: T_BNX; var a1, a2: T_BNX): Integer;
const
  {$IFDEF CPUX64}
  CBits  = 64;
  CHBits = 64 - 8;
  {$ELSE}
  CBits  = 32;
  CHBits = 32 - 8;
  {$ENDIF}

  procedure div_mod_(const A_a1, A_a2: T_BNX; ADivisor: UIntX; ADIndex: IntX);
  var
    Bu       : Boolean;
    V_xmul   : T_BNX;
    V_xdiv   : T_BNX;
    V_adiv   : T_BNX;
    V_adiu   : T_BNX;
    V_amod   : T_BNX;
    V_amou   : T_BNX;
    VDiv     : UIntX;
    VCompare : Integer;
  begin
    try
      //   
      Inc (Result);
      VDiv := BNX_div_intn (V_adiv, A_a1, ADivisor, ADIndex);
      BN_mul_intn (V_xmul, A_a2, V_adiv.Sign, VDiv, V_adiv.HIndex); // V_xmul = adiv * a2
      BN_sub (V_amod, A_a1, V_xmul); // amod = a1 - adiv * a2

      Bu := false;
      // V_adiv -  
      // V_amod -  

      while true do begin
        if Bu then begin
          // V_amou -  
          VCompare := BN_DataCompare (V_amou, A_a2);
        end else begin
          // V_amod -  
          VCompare := BN_DataCompare (V_amod, A_a2);
        end;

        if VCompare = 0 then begin
          //  
          //     
          if Bu then begin
            // V_adiu -  
            // V_amou -  
            if V_amou.Sign < 0 then begin
              BN_sub_int (adiv, V_adiu, 1, 1); // adiv = adiv - 1
            end else begin
              BN_add_int (adiv, V_adiu, 1, 1); // adiv = adiv + 1
            end;
          end else begin
            // V_adiv -  
            // V_amod -  
            if V_amod.Sign < 0 then begin
              BN_sub_int (adiv, V_adiv, 1, 1); // adiv = adiv - 1
            end else begin
              BN_add_int (adiv, V_adiv, 1, 1); // adiv = adiv + 1
            end;
          end;
          BN_AssignZero (amod);

          Exit;
        end;

        if VCompare < 0 then begin
          //  
          //     
          if Bu then begin
            // V_adiu -  
            // V_amou -  
            if V_amou.Sign < 0 then begin
              BN_sub_int (adiv, V_adiu, 1, 1); // adiv = adiv - 1
              BN_add (amod, V_amou, A_a2);     // amod = amod + a2
            end else begin
              BN_dup (adiv, V_adiu);
              BN_dup (amod, V_amou);
            end;
          end else begin
            // V_adiv -  
            // V_amod -  
            if V_amod.Sign < 0 then begin
              BN_sub_int (adiv, V_adiv, 1, 1); // adiv = adiv - 1
              BN_add (amod, V_amod, A_a2);     // amod = amod + a2
            end else begin
              BN_dup (adiv, V_adiv);
              BN_dup (amod, V_amod);
            end;
          end;

          Exit;
        end;

        //   
        Inc (Result);
        if Bu then begin
          // V_amou -  
          // V_amod -  
          VDiv := BNX_div_intn (V_xdiv, V_amou, ADivisor, ADIndex);
          BN_mul_intn (V_xmul, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // V_xmul = adiv * a2
          BN_sub (V_amod, V_amou, V_xmul); // amod = amod - adiv * a2

          // V_adiu -  
          // V_adiv -  
          BN_add (V_adiv, V_adiu, V_xdiv);
        end else begin
          // V_amod -  
          // V_amou -  
          VDiv := BNX_div_intn (V_xdiv, V_amod, ADivisor, ADIndex);
          BN_mul_intn (V_xmul, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // V_xmul = adiv * a2
          BN_sub (V_amou, V_amod, V_xmul); // amod = amod - adiv * a2

          // V_adiv -  
          // V_adiu -  
          BN_add (V_adiu, V_adiv, V_xdiv);
        end;
        Bu := not Bu;
      end;
    finally
      BN_Clear (V_xmul);
      BN_Clear (V_xdiv);
      BN_Clear (V_adiv);
      BN_Clear (V_adiu);
      BN_Clear (V_amod);
      BN_Clear (V_amou);
    end;
  end;

var
  V_a1     : T_BNX;
  V_a2     : T_BNX;
  VHBits   : UIntX;
  VShift   : IntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VSign    : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VData    : UIntX;
  VDIndex  : IntX;
  VDivisor : UIntX;
  VCompare : Integer;
begin
  Result := 0;

  VSign1 := a1.Sign;
  VSign2 := a2.Sign;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // |a2| = 0
    raise EDivByZero.Create (SDivByZero);
  end;

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // |a1| = 0
    // adiv = 0
    // amod = 0
    BN_AssignZero (adiv);
    BN_AssignZero (amod);
    Exit;
  end;

  VSign := BN_mul_div_sign (VSign1, VSign2);
  VCompare := BN_DataCompare (a1, a2);

  if VCompare = 0 then begin
    // |a1| = |a2|
    // adiv = 1
    // amod = 0
    BN_AssignInt (adiv, VSign, 1);
    BN_AssignZero (amod);
    Exit;
  end;

  if VCompare < 0 then begin
    // |a1| < |a2|
    // adiv = 0
    // amod = |a1|
    BN_AssignZero (adiv);
    BN_sdup (amod, 1, a1); // amod = |a1|
    Exit;
  end;

  // |a1| > |a2|

  VDIndex := VCount2 - 1;
  VDivisor := a2.Data[VDIndex];

  if VCount2 = 1 then begin
    VData := BN_div_mod_int (adiv, a1, 1, VDivisor);
    if VData = 0 then begin
      BN_AssignZero (amod);
    end else begin
      BN_AssignInt (amod, 1, VData);
    end;
    adiv.Sign := VSign;
    Exit;
  end;

  VHBits := BN_Bits (VDivisor);
  if VHBits < CHBits then begin
    //   VDivisor
    VShift := CBits - VHBits;
    BN_sdup (V_a1, 1, a1);
    BN_sdup (V_a2, 1, a2);
    BN_shl_(V_a1, VShift);
    BN_shl_(V_a2, VShift);
    VDivisor := V_a2.Data[VDIndex];

    div_mod_(V_a1, V_a2, VDivisor, VDIndex);
    BN_shr_(amod, VShift);

    BN_Clear (V_a1);
    BN_Clear (V_a2);
  end else begin
    a1.Sign := 1; // a1 = |a1|
    a2.Sign := 1; // a2 = |a2|

    div_mod_(a1, a2, VDivisor, VDIndex);

    a1.Sign := VSign1;
    a2.Sign := VSign2;
  end;
  adiv.Sign := VSign;
end;

{
  amod = a1 mod a2, amod > 0

  Result -  
  Result - number of iterations
}
function BN_mod (var amod: T_BNX; var a1, a2: T_BNX): Integer;
const
  {$IFDEF CPUX64}
  CBits  = 64;
  CHBits = 64 - 8;
  {$ELSE}
  CBits  = 32;
  CHBits = 32 - 8;
  {$ENDIF}

  procedure mod_(const A_a1, A_a2: T_BNX; ADivisor: UIntX; ADIndex: IntX);
  var
    Bu       : Boolean;
    V_xmul   : T_BNX;
    V_xdiv   : T_BNX;
    V_amod   : T_BNX;
    V_amou   : T_BNX;
    VDiv     : UIntX;
    VCompare : Integer;
  begin
    try
      //   
      Inc (Result);
      VDiv := BNX_div_intn (V_xdiv, A_a1, ADivisor, ADIndex);
      BN_mul_intn (V_xmul, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // V_xmul = adiv * a2
      BN_sub (V_amod, A_a1, V_xmul); // amod = a1 - adiv * a2

      Bu := false;
      // V_adiv -  
      // V_amod -  

      while true do begin
        if Bu then begin
          // V_amou -  
          VCompare := BN_DataCompare (V_amou, A_a2);
        end else begin
          // V_amod -  
          VCompare := BN_DataCompare (V_amod, A_a2);
        end;

        if VCompare = 0 then begin
          //  
          //     
          BN_AssignZero (amod);

          Exit;
        end;

        if VCompare < 0 then begin
          //  
          //     
          if Bu then begin
            // V_adiu -  
            // V_amou -  
            if V_amou.Sign < 0 then begin
              BN_add (amod, V_amou, A_a2); // amod = amod + a2
            end else begin
              BN_dup (amod, V_amou);
            end;
          end else begin
            // V_adiv -  
            // V_amod -  
            if V_amod.Sign < 0 then begin
              BN_add (amod, V_amod, A_a2); // amod = amod + a2
            end else begin
              BN_dup (amod, V_amod);
            end;
          end;

          Exit;
        end;

        //   
        Inc (Result);
        if Bu then begin
          // V_amou -  
          // V_amod -  
          VDiv := BNX_div_intn (V_xdiv, V_amou, ADivisor, ADIndex);
          BN_mul_intn (V_xmul, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // V_xmul = adiv * a2
          BN_sub (V_amod, V_amou, V_xmul); // amod = amod - adiv * a2
        end else begin
          // V_amod -  
          // V_amou -  
          VDiv := BNX_div_intn (V_xdiv, V_amod, ADivisor, ADIndex);
          BN_mul_intn (V_xmul, A_a2, V_xdiv.Sign, VDiv, V_xdiv.HIndex); // V_xmul = adiv * a2
          BN_sub (V_amou, V_amod, V_xmul); // amod = amod - adiv * a2
        end;
        Bu := not Bu;
      end;
    finally
      BN_Clear (V_xmul);
      BN_Clear (V_xdiv);
      BN_Clear (V_amod);
      BN_Clear (V_amou);
    end;
  end;

var
  V_a1     : T_BNX;
  V_a2     : T_BNX;
  VHBits   : UIntX;
  VShift   : IntX;
  VSign1   : ShortInt;
  VSign2   : ShortInt;
  VCount1  : IntX;
  VCount2  : IntX;
  VData    : UIntX;
  VDIndex  : IntX;
  VDivisor : UIntX;
  VCompare : Integer;
begin
  Result := 0;

  VSign1 := a1.Sign;
  VSign2 := a2.Sign;

  VCount1 := BN_DataCount (a1);
  VCount2 := BN_DataCount (a2);

  if (VSign2 = 0) or (VCount2 <= 0) then begin
    // |a2| = 0
    raise EDivByZero.Create (SDivByZero);
  end;

  if (VSign1 = 0) or (VCount1 <= 0) then begin
    // |a1| = 0
    // adiv = 0
    // amod = 0
    BN_AssignZero (amod);
    Exit;
  end;

  VCompare := BN_DataCompare (a1, a2);

  if VCompare = 0 then begin
    // |a1| = |a2|
    // adiv = 1
    // amod = 0
    BN_AssignZero (amod);
    Exit;
  end;

  if VCompare < 0 then begin
    // |a1| < |a2|
    // adiv = 0
    // amod = |a1|
    BN_sdup (amod, 1, a1); // amod = |a1|
    Exit;
  end;

  // |a1| > |a2|

  VDIndex := VCount2 - 1;
  VDivisor := a2.Data[VDIndex];

  if VCount2 = 1 then begin
    VData := BN_div_mod_int (V_a1, a1, 1, VDivisor);
    if VData = 0 then begin
      BN_AssignZero (amod);
    end else begin
      BN_AssignInt (amod, 1, VData);
    end;
    BN_Clear (V_a1);
    Exit;
  end;

  VHBits := BN_Bits (VDivisor);
  if VHBits < CHBits then begin
    //   VDivisor
    VShift := CBits - VHBits;
    BN_sdup (V_a1, 1, a1);
    BN_sdup (V_a2, 1, a2);
    BN_shl_(V_a1, VShift);
    BN_shl_(V_a2, VShift);
    VDivisor := V_a2.Data[VDIndex];

    mod_(V_a1, V_a2, VDivisor, VDIndex);
    BN_shr_(amod, VShift);

    BN_Clear (V_a1);
    BN_Clear (V_a2);
  end else begin
    a1.Sign := 1; // a1 = |a1|
    a2.Sign := 1; // a2 = |a2|

    mod_(a1, a2, VDivisor, VDIndex);

    a1.Sign := VSign1;
    a2.Sign := VSign2;
  end;
end;

{
  adiv = a1 div a2

  Result -  
  Result - number of iterations
}
function BN_div (var adiv: T_BNX; var a1, a2: T_BNX): Integer;
var
  amod : T_BNX;
begin
  Result := BN_div_mod (adiv, amod, a1, a2);
  BN_Clear (amod);
end;

{
  a1 div a2
  a1 mod a2
}
function BNX_div_mod_int_asm (a1: PUIntX; a2: UIntX; Count: IntX;
  adiv: PUIntX): UIntX;
{$IFDEF CPUX64}
// rcx | @a1   |
// rdx | a2    | r10
// r8  | Count |
// r9  | @adiv |
asm
        mov     r10, rdx
        xor     rdx, rdx
  @loop:
        mov     rax, [rcx]
        div     r10
        mov     qword ptr [r9], rax
        lea     rcx, [rcx - 8]
        lea     r9, [r9 - 8]
     // sub     rcx, 8
     // sub     r9, 8
  @next:
        dec     r8
        jnz     @loop

        mov     rax, rdx
end;
{$ELSE}
// eax | @a1   | esi
// edx | a2    | edi
// ecx | Count |
//     | @adiv | ebx
asm
        push    ebx
        push    esi
        push    edi

        mov     esi, eax
        mov     ebx, [adiv]
        mov     edi, edx
        xor     edx, edx
  @loop:
        mov     eax, [esi]
        div     edi
        mov     dword ptr [ebx], eax
        lea     esi, [esi - 4]
        lea     ebx, [ebx - 4]
     // sub     esi, 4
     // sub     ebx, 4
  @next:
        dec     ecx
        jnz     @loop

        mov     eax, edx

        pop     edi
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
  adiv = a1 div a2
}
procedure BN_div_int (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX);
begin
  BN_div_mod_int (adiv, a1, ASign2, a2);
end;

{
  adiv = a1 div a2
  a2 - signed integer
}
procedure BN_div_sint (var adiv: T_BNX; const a1: T_BNX; a2: IntX);
begin
  BN_div_mod_sint (adiv, a1, a2);
end;

{
  adiv = a1 div a2'
  a2'[n] = a2,   (other elements are zeros)
}
procedure BN_div_intn (var adiv: T_BNX; const a1: T_BNX; ASign2: ShortInt; a2: UIntX;
  n: IntX);
begin
  BN_div_mod_intn (adiv, a1, ASign2, a2, n);
end;

{
  adiv = a1 div a2'
  a2'[n] = a2,   (other elements are zeros)
  a2 - signed integer
}
procedure BN_div_sintn (var adiv: T_BNX; const a1: T_BNX; a2: IntX;
  n: IntX);
begin
  BN_div_mod_sintn (adiv, a1, a2, n);
end;

{
  adiv = a1 div a2
  amod = a1 mod a2, amod > 0
  Result = amod
}
function BN_div_mod_int (var adiv: T_BNX; const a1: T_BNX;
  ASign2: ShortInt; a2: UIntX): UIntX;
var
  VSign  : ShortInt;
  VCount : IntX;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  Result := 0;
  VSign := BN_mul_div_sign (a1.Sign, ASign2);
  VCount := BN_Init (adiv, VSign, a1.Count);
  if VCount <= 0 then Exit;

  Result := BNX_div_mod_int_asm (a1.PHData, a2, VCount, adiv.PHData);
  BN_Pack (adiv);
end;

{
  adiv = a1 div a2
  amod = a1 mod a2, amod > 0
  a2 - signed integer
  Result = amod
}
function BN_div_mod_sint (var adiv: T_BNX; const a1: T_BNX; a2: IntX): UIntX;
var
  VSign2 : ShortInt;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  if a2 > 0 then VSign2 := 1
  else begin
    VSign2 := -1;
    a2 := -a2;
  end;

  Result := BN_div_mod_int (adiv, a1, VSign2, a2);
end;

{
  adiv = a1 div a2'
  amod = a1 mod a2', amod > 0
  a2'[n] = a2,   (other elements are zeros)
  Result = amod
}
function BN_div_mod_intn (var adiv: T_BNX; const a1: T_BNX;
  ASign2: ShortInt; a2: UIntX; n: IntX): UIntX;
var
  VSign  : ShortInt;
  VCount : IntX;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  Result := 0;
  VSign := BN_mul_div_sign (a1.Sign, ASign2);
  VCount := BN_Init (adiv, VSign, a1.Count - n);
  if VCount <= 0 then Exit;

  Result := BNX_div_mod_int_asm (a1.PHData, a2, VCount, adiv.PHData);
  BN_Pack (adiv);
end;

{
  adiv = a1 div a2'
  amod = a1 mod a2', amod > 0
  a2'[n] = a2,   (other elements are zeros)
  a2 - signed integer
  Result = amod
}
function BN_div_mod_sintn (var adiv: T_BNX; const a1: T_BNX; a2: IntX;
  n: IntX): UIntX;
var
  VSign2 : ShortInt;
begin
  if a2 = 0 then begin
    raise EDivByZero.Create (SDivByZero);
  end;

  if a2 > 0 then VSign2 := 1
  else begin
    VSign2 := -1;
    a2 := -a2;
  end;

  Result := BN_div_mod_intn (adiv, a1, VSign2, a2, n);
end;

{
  P = ABase^AExp
}
procedure BN_exp_base (var P: T_BNX; const ABase: T_BNX; AExp: UIntX);
var
  V     : T_BNX;
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_dup (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignInt (P, 1, 1); // P = 1
  if AExp = 0 then Exit;

  BN_dup (VBase, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_mul_base (V, VBase, VBase);
      BN_dup (VBase, V); // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_dup (V, P);
    BN_mul_base (P, V, VBase); // P = P * VBase
  end;

  BN_Clear (V);
  BN_Clear (VBase);
end;

{
  P = ABase^AExp
}
procedure BN_exp (var P: T_BNX; const ABase: T_BNX; AExp: UIntX);
var
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_dup (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignInt (P, 1, 1); // P = 1
  if AExp = 0 then Exit;

  BN_dup (VBase, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase); // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;

  BN_Clear (VBase);
end;

{
  P = ABase^AExp
}
procedure BN_exp_int (var P: T_BNX; ASign: ShortInt; ABase: UIntX; AExp: UIntX);
var
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_AssignInt (P, ASign, ABase); // P = ABase
    Exit;
  end;

  BN_AssignInt (P, 1, 1); // P = 1
  if AExp = 0 then Exit;

  BN_AssignInt (VBase, ASign, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase); // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;

  BN_Clear (VBase);
end;

{
  P = ABase^AExp
  ABase - signed integer
}
procedure BN_exp_sint (var P: T_BNX; ABase: IntX; AExp: UIntX);
var
  VBase : T_BNX;
begin
  if AExp = 1 then begin
    BN_AssignSInt (P, ABase); // P = ABase
    Exit;
  end;

  BN_AssignInt (P, 1, 1); // P = 1
  if AExp = 0 then Exit;

  BN_AssignSInt (VBase, ABase);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase); // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;

  BN_Clear (VBase);
end;

{
  P = 10^AExp = 5^AExp * 2^AExp
}
procedure BN_10exp (var P: T_BNX; ASign: ShortInt; AExp: UIntX);
var
  VBase : T_BNX;
  VExp  : UIntX;
begin
  if AExp = 1 then begin
    BN_AssignInt (P, ASign, 10); // P = 10
    Exit;
  end;

  BN_AssignInt (P, 1, 1); // P = 1
  if AExp = 0 then Exit;

  VExp := AExp;
  BN_AssignInt (VBase, ASign, 5);
  while AExp <> 0 do begin
    while not Odd (AExp) do begin
      AExp := AExp shr 1;
      BN_sqr_(VBase); // VBase = VBase * VBase
    end;
    Dec (AExp);
    BN_mul_(P, VBase); // P = P * VBase
  end;
  BN_shl_(P, VExp); // P = P * 2^Exp

  BN_Clear (VBase);
end;

{
  P = ABase^2
}
procedure BN_sqr_base (var P: T_BNX; const ABase: T_BNX);
var
  V : T_BNX;
begin
  BN_mul_base (V, ABase, ABase); // V = ABase^2
  BN_dup (P, V);                 // P = ABase^2
  BN_Clear (V);
end;

{
  ABase = ABase^2
}
procedure BN_sqr_base_(var ABase: T_BNX);
var
  V : T_BNX;
begin
  BN_mul_base (V, ABase, ABase); // V = ABase^2
  BN_dup (ABase, V);             // ABase = ABase^2
  BN_Clear (V);
end;

{
  P = ABase^2
    
  Using the Karatsuba algorithm
}
procedure BN_sqr (var P: T_BNX; const ABase: T_BNX);
var
  VSignP  : ShortInt;
  VCountB : IntX;
  Vm      : IntX;

  V_a0    : T_BNX;
  V_a1    : T_BNX;
  V_aa    : T_BNX;
  V_aa00  : T_BNX;
  V_aa11  : T_BNX;
  V_aaaa  : T_BNX;
begin
  if ABase.Sign = 0 then VSignP := 0 else VSignP := 1;
  VCountB := BN_DataCount (ABase);

  if (VSignP = 0) or (VCountB <= 0) then begin
    BN_AssignZero (P);
    Exit;
  end;

  if VCountB < Karatsuba_sqr_TCount then begin
    BN_sqr_base (P, ABase);
    Exit;
  end;

  Vm := VCountB shr 1; // VCountB div 2;

  // (a0 + a1 * 2^[m]) * (a0 + a1 * 2^[m]) =
  // = (a0 * a0) + (2 * a0 * a1) * 2^[m] + (a1 * a1) * 2^[2m]
  // | aa00 = (a0 * a0)
  // | aa11 = (a1 * a1)
  // | aaaa = (a0 + a1) * (a0 + a1) = (a0 * a0) + (a1 * a1) + (2 * a0 * a1)
  // | (2 * a0 * a1) = (a0 + a1) * (a0 + a1) - ((a0 * a0) + (a1 * a1))
  // | = aaaa - (aa00 + aa11)
  // = aa00 + (aaaa - (aa00 + aa11)) *  2^[m] + aa11 * 2^[2m]

  BN_cdup (V_a0, ABase, Vm);
  BN_hdup (V_a1, ABase, Vm);

  BN_add (V_aa, V_a0, V_a1);     // V_aa = V_a0 + V_a1

  BN_sqr (V_aa00, V_a0);         // V_aa00 = a0 * a0
  BN_sqr (V_aa11, V_a1);         // V_aa11 = a1 * a1
  BN_sqr (V_aaaa, V_aa);         // V_aaaa = (a0 + a1) * (a0 + a1)

  BN_add (V_a0, V_aa00, V_aa11); // V_a0 = V_aa00 + V_aa11
  BN_sub (V_aa, V_aaaa, V_a0);   // V_aa = V_aaaa - (V_aa00 + V_aa11)
  BN_cshl_(V_aa, Vm);            // V_aa = (V_aaaa - (V_aa00 + V_aa11)) * 2^[m]
  BN_cshl_(V_aa11, Vm shl 1);    // V_aa11 = V_aa11 * 2^[2m]
  BN_add (V_a0, V_aa00, V_aa);   // V_a0 = V_aa00 + (V_aaaa - (V_aa00 + V_aa11)) * 2^[m]
  BN_add (P, V_a0, V_aa11);      // P = V_aa00 + (V_aaaa - (V_aa00 + V_aa11)) * 2^[m] + V_aa11 * 2^[2m]
  P.Sign := VSignP;

  BN_Clear (V_a0);
  BN_Clear (V_a1);
  BN_Clear (V_aa);
  BN_Clear (V_aa00);
  BN_Clear (V_aa11);
  BN_Clear (V_aaaa);
end;

{
  ABase = ABase^2
    
  Using the Karatsuba algorithm
}
procedure BN_sqr_(var ABase: T_BNX);
var
  VSignP  : ShortInt;
  VCountB : IntX;
  Vm      : IntX;

  V_a0    : T_BNX;
  V_a1    : T_BNX;
  V_aa    : T_BNX;
  V_aa00  : T_BNX;
  V_aa11  : T_BNX;
  V_aaaa  : T_BNX;
begin
  if ABase.Sign = 0 then VSignP := 0 else VSignP := 1;
  VCountB := BN_DataCount (ABase);

  if (VSignP = 0) or (VCountB <= 0) then begin
    BN_AssignZero (ABase);
    Exit;
  end;

  if VCountB < Karatsuba_sqr_TCount then begin
    BN_sqr_base_(ABase);
    Exit;
  end;

  Vm := VCountB shr 1; // VCountB div 2;

  // (a0 + a1 * 2^[m]) * (a0 + a1 * 2^[m]) =
  // = (a0 * a0) + (2 * a0 * a1) * 2^[m] + (a1 * a1) * 2^[2m]
  // | aa00 = (a0 * a0)
  // | aa11 = (a1 * a1)
  // | aaaa = (a0 + a1) * (a0 + a1) = (a0 * a0) + (a1 * a1) + (2 * a0 * a1)
  // | (2 * a0 * a1) = (a0 + a1) * (a0 + a1) - ((a0 * a0) + (a1 * a1))
  // | = aaaa - (aa00 + aa11)
  // = aa00 + (aaaa - (aa00 + aa11)) *  2^[m] + aa11 * 2^[2m]

  BN_cdup (V_a0, ABase, Vm);
  BN_hdup (V_a1, ABase, Vm);

  BN_add (V_aa, V_a0, V_a1);     // V_aa = V_a0 + V_a1

  BN_sqr (V_aa00, V_a0);         // V_aa00 = a0 * a0
  BN_sqr (V_aa11, V_a1);         // V_aa11 = a1 * a1
  BN_sqr (V_aaaa, V_aa);         // V_aaaa = (a0 + a1) * (a0 + a1)

  BN_add (V_a0, V_aa00, V_aa11); // V_a0 = V_aa00 + V_aa11
  BN_sub (V_aa, V_aaaa, V_a0);   // V_aa = V_aaaa - (V_aa00 + V_aa11)
  BN_cshl_(V_aa, Vm);            // V_aa = (V_aaaa - (V_aa00 + V_aa11)) * 2^[m]
  BN_cshl_(V_aa11, Vm shl 1);    // V_aa11 = V_aa11 * 2^[2m]
  BN_add (V_a0, V_aa00, V_aa);   // V_a0 = V_aa00 + (V_aaaa - (V_aa00 + V_aa11)) * 2^[m]
  BN_add (ABase, V_a0, V_aa11);  // ABase = V_aa00 + (V_aaaa - (V_aa00 + V_aa11)) * 2^[m] + V_aa11 * 2^[2m]
  ABase.Sign := VSignP;

  BN_Clear (V_a0);
  BN_Clear (V_a1);
  BN_Clear (V_aa);
  BN_Clear (V_aa00);
  BN_Clear (V_aa11);
  BN_Clear (V_aaaa);
end;

{
  P = ABase^3
}
procedure BN_exp3 (var P: T_BNX; const ABase: T_BNX);
var
  V2 : T_BNX;
begin
  BN_sqr (V2, ABase);    // V2 = ABase^2
  BN_mul (P, V2, ABase); // P  = ABase^3
  BN_Clear (V2);
end;

{
  P = ABase^4
}
procedure BN_exp4 (var P: T_BNX; const ABase: T_BNX);
var
  V2 : T_BNX;
begin
  BN_sqr (V2, ABase); // V2 = ABase^2
  BN_sqr (P, V2);     // P  = ABase^4
  BN_Clear (V2);
end;

{
  P = ABase^5
}
procedure BN_exp5 (var P: T_BNX; const ABase: T_BNX);
var
  V2, V4 : T_BNX;
begin
  BN_sqr (V2, ABase);    // V2 = ABase^2
  BN_sqr (V4, V2);       // V4 = ABase^4
  BN_mul (P, V4, ABase); // P  = ABase^5
  BN_Clear (V2);
  BN_Clear (V4);
end;

{
  Internal procedure

   n- 
   

  n-th Root
  Newton's method

  aroot = |a|^(1/n)

  Result -  
  Result - number of iterations
}
function BN_root_(var aroot: T_BNX; var a: T_BNX; n: UIntX): Integer;
var
  B_vroot  : Boolean;
  Vnm      : UIntX;
  vroot    : T_BNX;
  VX, VXS  : T_BNX;
  VXN, VNX : T_BNX;
  VCompare : Integer;
begin
  Result := 0;

  try

  Vnm := n - 1;

  //   
  Inc (Result);
  // x0 = ([n - 1] + a) / n
  BN_add_int (VX, a, 1, Vnm);   // VX = [n - 1] + a
  BN_div_int (vroot, VX, 1, n); // vroot = ([n - 1] + a) / n
  (*
  // x0 = a
  BN_dup (vroot, a); // vroot = a
  // *)

  B_vroot := true;
  // vroot -  

  while true do begin
    //   
    Inc (Result);
    // x1 = ([n - 1] * x0 + a / x0^[n - 1]) / n
    if B_vroot then begin
      // vroot -  
      // aroot -  
      BN_exp (VXN, vroot, Vnm);        // VXN = x0^[n - 1]
      BN_div (VX, a, VXN);             // VX = a / x0^[n - 1]
      BN_mul_int (VNX, vroot, 1, Vnm); // VNX = [n - 1] * x0
      BN_add (VXS, VNX, VX);           // VXS = [n - 1] * x0 + a / x0^[n - 1]
      BN_div_int (aroot, VXS, 1, n);   // aroot = ([n - 1] * x0 + a / x0^[n - 1]) / n
    end else begin
      // aroot -  
      // vroot -  
      BN_exp (VXN, aroot, Vnm);        // VXN = x0^[n - 1]
      BN_div (VX, a, VXN);             // VX = a / x0^[n - 1]
      BN_mul_int (VNX, aroot, 1, Vnm); // VNX = [n - 1] * x0
      BN_add (VXS, VNX, VX);           // VXS = [n - 1] * x0 + a / x0^[n - 1]
      BN_div_int (vroot, VXS, 1, n);   // aroot = ([n - 1] * x0 + a / x0^[n - 1]) / n
    end;
    B_vroot := not B_vroot;

    if B_vroot then begin
      // vroot -  
      // aroot -  
      VCompare := BN_DataCompare (vroot, aroot);
    end else begin
      // aroot -  
      // vroot -  
      VCompare := BN_DataCompare (aroot, vroot);
    end;

    if VCompare > 0 then begin
      //  
      //   
      if B_vroot then begin
        // vroot -  
        // aroot -  
        // BN_dup (vroot, aroot); // aroot = vroot
      end else begin
        // aroot -  
        // vroot -  
        BN_dup (aroot, vroot); // aroot = vroot
      end;
      // VCompare := 0;
      Exit;
    end;
    if VCompare = 0 then begin
      //  
      //    
      Exit;
    end;
  end;

  finally
    BN_Clear (vroot);
    BN_Clear (VX);
    BN_Clear (VXS);
    BN_Clear (VXN);
    BN_Clear (VNX);
  end;
end;

{
   n- 
   

  n-th Root
  Newton's method

  aroot = |a|^(1/n)

  Result -  
  Result - number of iterations
}
function BN_root (var aroot: T_BNX; var a: T_BNX; n: UIntX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) or (n = 0) then begin
    // a = 0
    // aroot = 0
    BN_AssignZero (aroot);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    BN_AssignInt (aroot, 1, 1);
    Exit;
  end;

  if n = 1 then begin
    // aroot = |a|
    BN_sdup (aroot, 1, a);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_root_(aroot, a, n);

  a.Sign := VSign;
end;

{
   n- 
   

  n-th Root
  Newton's method

  aroot = |a|^(1/n)
  amod = |a| - aroot^n

  Result -  
  Result - number of iterations
}
function BN_root_mod (var aroot, amod: T_BNX; var a: T_BNX; n: UIntX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
  VP     : T_BNX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) or (n = 0) then begin
    // a = 0
    // aroot = 0
    // amod = 0
    BN_AssignZero (aroot);
    BN_AssignZero (amod);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    // amod = 0
    BN_AssignInt (aroot, 1, 1);
    BN_AssignZero (amod);
    Exit;
  end;

  if n = 1 then begin
    // aroot = |a|
    // amod = 0
    BN_sdup (aroot, 1, a);
    BN_AssignZero (amod);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_root_(aroot, a, n);
  BN_exp (VP, aroot, n); // VP = aroot^n
  BN_sub (amod, a, VP);  // amod = |a| - aroot^n
  BN_Clear (VP);

  a.Sign := VSign;
end;

{
  Internal procedure

   
   
    

  Square Root
  Newton's method

  a > 0
  aroot = |a|^(1/2)
  vroot = aroot

  Result -  
  Result - number of iterations
}
function BN_sqrt_(var aroot, vroot: T_BNX; var a: T_BNX): Integer;
var
  B_vroot  : Boolean;
  VX       : T_BNX;
  VCompare : Integer;
begin
  Result := 0;

  try

  //   
  Inc (Result);
  // x0 = (1 + a) / 2
  BN_add_int (vroot, a, 1, 1); // vroot = 1 + a
  BN_shr_(vroot, 1);           // vroot = (1 + a) / 2
  (*
  // x0 = a
  BN_dup (vroot, a); // vroot = a
  // *)

  B_vroot := true;
  // vroot -  

  while true do begin
    //   
    Inc (Result);
    // x1 = (x0 + a / x0) / 2
    if B_vroot then begin
      // vroot -  
      // aroot -  
      BN_div (VX, a, vroot);     // VX = a / x0
      BN_add (aroot, vroot, VX); // aroot = x0 + a / x0
      BN_shr_(aroot, 1);         // aroot = (x0 + a / x0) / 2
    end else begin
      // aroot -  
      // vroot -  
      BN_div (VX, a, aroot);     // VX = a / x0
      BN_add (vroot, aroot, VX); // aroot = x0 + a / x0
      BN_shr_(vroot, 1);         // aroot = (x0 + a / x0) / 2
    end;
    B_vroot := not B_vroot;

    if B_vroot then begin
      // vroot -  
      // aroot -  
      VCompare := BN_DataCompare (vroot, aroot);
    end else begin
      // aroot -  
      // vroot -  
      VCompare := BN_DataCompare (aroot, vroot);
    end;

    if VCompare > 0 then begin
      //  
      //   
      if B_vroot then begin
        // vroot -  
        // aroot -  
        BN_dup (vroot, aroot); // aroot = vroot
      end else begin
        // aroot -  
        // vroot -  
        BN_dup (aroot, vroot); // aroot = vroot
      end;
      // VCompare := 0;
      Exit;
    end;
    if VCompare = 0 then begin
      //  
      //    
      Exit;
    end;
  end;

  finally
    BN_Clear (VX);
  end;
end;

{
   
   
    

  Square Root
  Newton's method

  aroot = |a|^(1/2)

  Result -  
  Result - number of iterations
}
function BN_sqrt (var aroot: T_BNX; var a: T_BNX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
  vroot  : T_BNX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) then begin
    // a = 0
    // aroot = 0
    BN_AssignZero (aroot);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    BN_AssignInt (aroot, 1, 1);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_sqrt_(aroot, vroot, a);
  BN_Clear (vroot);

  a.Sign := VSign;
end;

{
   
   
    

  Square Root
  Newton's method

  aroot = |a|^(1/2)
  amod = |a| - aroot^2

  Result -  
  Result - number of iterations
}
function BN_sqrt_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
  vroot  : T_BNX;
  VX2    : T_BNX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) then begin
    // a = 0
    // aroot = 0
    // amod = 0
    BN_AssignZero (aroot);
    BN_AssignZero (amod);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    // amod = 0
    BN_AssignInt (aroot, 1, 1);
    BN_AssignZero (amod);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_sqrt_(aroot, vroot, a);
  BN_mul (VX2, aroot, vroot); // VX2 = aroot^2
  BN_sub (amod, a, VX2);      // amod = a - aroot^2
  BN_Clear (vroot);
  BN_Clear (VX2);

  a.Sign := VSign;
end;

{
  Internal procedure

   
   

  Cube Root
  Newton's method

  a > 0
  aroot = |a|^(1/3)
  vroot = aroot

  Result -  
  Result - number of iterations
}
function BN_root3_(var aroot, vroot: T_BNX; var a: T_BNX): Integer;
var
  VX, VXS  : T_BNX;
  V2X, VX2 : T_BNX;
  VCompare : Integer;
begin
  Result := 0;

  try

  //   
  Inc (Result);
  // x0 = (2 + a) / 3
  BN_add_int (VX, a, 1, 2);     // VX = 2 + a
  BN_div_int (vroot, VX, 1, 3); // vroot = (2 + a) / 3
  (*
  // x0 = a
  BN_dup (vroot, a); // vroot = a
  // *)

  while true do begin
    //   
    Inc (Result);
    // x1 = (2 * x0 + a / x0^2) / 3
    BN_dup (VX, vroot);            // VX = x0
    BN_mul (VX2, vroot, VX);       // VX2 = x0^2
    BN_div (VX, a, VX2);           // VX = a / x0^2
    BN_shl_(V2X, vroot, 1);        // V2X = 2 * x0
    BN_add (VXS, V2X, VX);         // VXS = 2 * x0 + a / x0^2
    BN_div_int (aroot, VXS, 1, 3); // aroot = (2 * x0 + a / x0^2) / 3

    // aroot -  
    // vroot -  
    VCompare := BN_DataCompare (aroot, vroot);

    if VCompare > 0 then begin
      //  
      //   
      BN_dup (aroot, vroot); // aroot = vroot
      // VCompare := 0;
      Exit;
    end;
    if VCompare = 0 then begin
      //  
      //    
      Exit;
    end;

    BN_dup (vroot, aroot);
    // aroot -  
    // vroot -  
  end;

  finally
    BN_Clear (VX);
    BN_Clear (VXS);
    BN_Clear (V2X);
    BN_Clear (VX2);
  end;
end;

{
   
   

  Cube Root
  Newton's method

  aroot = |a|^(1/3)

  Result -  
  Result - number of iterations
}
function BN_root3 (var aroot: T_BNX; var a: T_BNX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
  vroot  : T_BNX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) then begin
    // a = 0
    // aroot = 0
    BN_AssignZero (aroot);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    BN_AssignInt (aroot, 1, 1);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_root3_(aroot, vroot, a);
  BN_Clear (vroot);

  a.Sign := VSign;
end;

{
   
   

  Cube Root
  Newton's method

  aroot = |a|^(1/3)
  amod = |a| - aroot^3

  Result -  
  Result - number of iterations
}
function BN_root3_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
  vroot  : T_BNX;
  VX2    : T_BNX;
  VX3    : T_BNX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) then begin
    // a = 0
    // aroot = 0
    // amod = 0
    BN_AssignZero (aroot);
    BN_AssignZero (amod);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    // amod = 0
    BN_AssignInt (aroot, 1, 1);
    BN_AssignZero (amod);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_root3_(aroot, vroot, a);
  BN_mul (VX2, aroot, vroot); // VX2 = aroot^2
  BN_mul (VX3, aroot, VX2);   // VX3 = aroot^3
  BN_sub (amod, a, VX3);      // amod = a - aroot^3
  BN_Clear (vroot);
  BN_Clear (VX2);
  BN_Clear (VX3);

  a.Sign := VSign;
end;

{
   4- 
   

  4-th Root
  Newton's method

  aroot = |a|^(1/4)

  Result -  
  Result - number of iterations
}
function BN_root4 (var aroot: T_BNX; var a: T_BNX): Integer;
var
  R2 : Integer;
  VR : T_BNX;
begin
  R2 := BN_sqrt (VR, a);
  Result := R2 + BN_sqrt (aroot, VR);
  BN_Clear (VR);
end;

{
   4- 
   

  4-th Root
  Newton's method

  aroot = |a|^(1/4)
  amod = |a| - aroot^4

  Result -  
  Result - number of iterations
}
function BN_root4_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;
var
  R2    : Integer;
  VR    : T_BNX;
  VP    : T_BNX;
  VSign : ShortInt;
begin
  R2 := BN_sqrt (VR, a);
  Result := R2 + BN_sqrt (aroot, VR);
  BN_Clear (VR);

  VSign := a.Sign;
  a.Sign := 1; // a = |a|

  BN_exp4 (VP, aroot);  // VP = aroot^4
  BN_sub (amod, a, VP); // amod = |a| - aroot^4
  BN_Clear (VP);

  a.Sign := VSign;
end;

{
  Internal procedure

   5- 
   

  5-th Root
  Newton's method

  a > 0
  aroot = |a|^(1/5)
  vroot = aroot

  Result -  
  Result - number of iterations
}
function BN_root5_(var aroot, vroot: T_BNX; var a: T_BNX): Integer;
var
  VX, VX2  : T_BNX;
  VX4, V4X : T_BNX;
  VXS      : T_BNX;
  VCompare : Integer;
begin
  Result := 0;

  try

  //   
  Inc (Result);
  // x0 = (4 + a) / 5
  BN_add_int (VX, a, 1, 4);     // VX = 4 + a
  BN_div_int (vroot, VX, 1, 5); // vroot = (4 + a) / 5
  (*
  // x0 = a
  BN_dup (vroot, a); // vroot = a
  // *)

  while true do begin
    //   
    Inc (Result);
    // x1 = (4 * x0 + a / x0^4) / 5
    BN_dup (VX, vroot);            // VX = x0
    BN_mul (VX2, vroot, VX);       // VX2 = x0^2
    BN_dup (VX, VX2);              // VX = x0^2
    BN_mul (VX4, VX2, VX);         // VX4 = x0^4
    BN_div (VX, a, VX4);           // VX = a / x0^4
    BN_shl_(V4X, vroot, 2);        // V4X = 4 * x0
    BN_add (VXS, V4X, VX);         // VXS = 4 * x0 + a / x0^4
    BN_div_int (aroot, VXS, 1, 5); // aroot = (4 * x0 + a / x0^4) / 5

    // aroot -  
    // vroot -  
    VCompare := BN_DataCompare (aroot, vroot);

    if VCompare > 0 then begin
      //  
      //   
      BN_dup (aroot, vroot); // aroot = vroot
      // VCompare := 0;
      Exit;
    end;
    if VCompare = 0 then begin
      //  
      //    
      Exit;
    end;

    BN_dup (vroot, aroot);
    // aroot -  
    // vroot -  
  end;

  finally
    BN_Clear (VX);
    BN_Clear (VX2);
    BN_Clear (VX4);
    BN_Clear (V4X);
    BN_Clear (VXS);
  end;
end;

{
   5- 
   

  5-th Root
  Newton's method

  aroot = |a|^(1/5)

  Result -  
  Result - number of iterations
}
function BN_root5 (var aroot: T_BNX; var a: T_BNX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
  vroot  : T_BNX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) then begin
    // a = 0
    // aroot = 0
    BN_AssignZero (aroot);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    BN_AssignInt (aroot, 1, 1);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_root5_(aroot, vroot, a);
  BN_Clear (vroot);

  a.Sign := VSign;
end;

{
   5- 
   

  5-th Root
  Newton's method

  aroot = |a|^(1/5)
  amod = |a| - aroot^5

  Result -  
  Result - number of iterations
}
function BN_root5_mod (var aroot, amod: T_BNX; var a: T_BNX): Integer;
var
  VSign  : ShortInt;
  VCount : IntX;
  vroot  : T_BNX;
  VX2    : T_BNX;
  VY2    : T_BNX;
  VX4    : T_BNX;
  VX5    : T_BNX;
begin
  Result := 0;

  VSign := a.Sign;
  VCount := BN_DataCount (a);

  if (VSign = 0) or (VCount <= 0) then begin
    // a = 0
    // aroot = 0
    // amod = 0
    BN_AssignZero (aroot);
    BN_AssignZero (amod);
    Exit;
  end;

  if BN_DataValue (a) = 1 then begin
    // |a| = 1
    // aroot = 1
    // amod = 0
    BN_AssignInt (aroot, 1, 1);
    BN_AssignZero (amod);
    Exit;
  end;

  a.Sign := 1; // a = |a|

  Result := BN_root5_(aroot, vroot, a);
  BN_mul (VX2, aroot, vroot); // VX2 = aroot^2
  BN_dup (VY2, VX2);          // VY2 = aroot^2
  BN_mul (VX4, VY2, VX2);     // VX4 = aroot^4
  BN_mul (VX5, aroot, VX4);   // VX5 = aroot^5
  BN_sub (amod, a, VX5);      // amod = a - aroot^5
  BN_Clear (vroot);
  BN_Clear (VX2);
  BN_Clear (VY2);
  BN_Clear (VX4);
  BN_Clear (VX5);

  a.Sign := VSign;
end;

{
  a = a' * 2^n
  Result = n

  n =    
  n = number of low order zero bits
}
function BN_m2n (const a: T_BNX): UIntX;
var
  I : LoopInt;
  V : UIntX;
  O : Boolean;
begin
  Result := 0;

  for I := 0 to a.Count - 1 do begin
    V := a.Data[I];
    if V = 0 then begin
      {$IFDEF CPUX64}
      Inc (Result, 64);
      {$ELSE}
      Inc (Result, 32);
      {$ENDIF}
    end else begin
      O := Odd (V);
      if O then Break;
      repeat
        Inc (Result);
        V := V shr 1;
        O := Odd (V);
      until O;
      if O then Break;
    end;
  end;
end;

{
  a = a' * 2^n
  Result = n

  n =    
  n = number of low order zero bits
}
function BN_m2n_(const a: T_BNX; var sha: T_BNX): UIntX;
begin
  Result := BN_m2n (a);
  if Result <> 0 then BN_shr_(sha, a, Result);
end;

{
  a = a' * 2^n
  Result = n

  n =    
  n = number of low order zero bits
}
function BN_m2n_int (a: UIntX): UIntX;
begin
  Result := 0;
  if a = 0 then Exit;

  while not Odd (a) do begin
    Inc (Result);
    a := a shr 1;
  end;
end;

{
  a = a' * 2^n
  Result = n

  n =    
  n = number of low order zero bits
}
function BN_m2n_int_(a: UIntX; var sha: UIntX): UIntX;
begin
  Result := 0;
  sha := a;
  if sha = 0 then Exit;

  while not Odd (sha) do begin
    Inc (Result);
    sha := sha shr 1;
  end;
end;

{
    
  Greatest common divisor
  Highest common factor

   
  Euclidean algorithm

  (m,n)
  (m,n)
  gcd(m,n)
  hcf(m,n)
}
procedure BN_gcd (var R: T_BNX; var m, n: T_BNX);
var
  VSign_m  : ShortInt;
  VSign_n  : ShortInt;
  VUInt    : UIntX;
  VR       : array [0..2] of T_BNX;
  VI       : array [0..2] of Integer;
  VI0      : Integer;
  VCompare : Integer;
begin
  VSign_m := m.Sign;
  VSign_n := n.Sign;

  m.Sign := 1;
  n.Sign := 1;
  try
    // (0, n) = n
    // (1, n) = 1
    // (m, 0) = m
    // (m, 1) = 1
    // (m, m) = m
    // (-m, n) = (m, n)

    if BN_IsInt (m, VUInt) then begin
      if VUInt = 0 then begin // m = 0
        BN_dup (R, n);        //  (0, n) = n
        Exit;
      end;
      if VUInt = 1 then begin   // m = 1
        BN_AssignInt (R, 1, 1); //  (1, n) = 1
        Exit;
      end;
    end;

    if BN_IsInt (n, VUInt) then begin
      if VUInt = 0 then begin // n = 0
        BN_dup (R, m);        //  (m, 0) = m
        Exit;
      end;
      if VUInt = 1 then begin   // n = 1
        BN_AssignInt (R, 1, 1); //  (m, 1) = 1
        Exit;
      end;
    end;

    VCompare := BN_DataCompare (m, n);

    if VCompare = 0 then begin
      BN_dup (R, m); //  (m, m) = m
      Exit;
    end;

    if VCompare > 0 then begin
      // m > n
      BN_dup (VR[0], m);
      BN_dup (VR[1], n);
    end else begin
      // m < n
      BN_dup (VR[0], n);
      BN_dup (VR[1], m);
    end;

    VI[0] := 0;
    VI[1] := 1;
    VI[2] := 2;

    while true do begin
      BN_mod (VR[VI[2]], VR[VI[0]], VR[VI[1]]);
      if BN_IsZero (VR[VI[2]]) then begin
        BN_dup (R, VR[VI[1]]);
        Exit;
      end;

      VI0 := VI[0];
      VI[0] := VI[1];
      VI[1] := VI[2];
      VI[2] := VI0;
    end;
  finally
    m.Sign := VSign_m;
    n.Sign := VSign_n;

    BN_Clear (VR[0]);
    BN_Clear (VR[1]);
    BN_Clear (VR[2]);
  end;
end;

{
    
  Greatest common divisor
  Highest common factor

   
  Binary algorithm

  !!!  
  !!!      - Stack overflow
  !!!     

  !!! Cannot be used
  !!! Does not work with large numbers - Stack overflow
  !!! Works slowly with small numbers

  (m,n)
  (m,n)
  gcd(m,n)
  hcf(m,n)
}
procedure BN_gcd_bin (var R: T_BNX; var m, n: T_BNX);
var
  VSign_m  : ShortInt;
  VSign_n  : ShortInt;
  Vm_odd   : Boolean;
  Vm_even  : Boolean;
  Vn_odd   : Boolean;
  Vn_even  : Boolean;
  VUInt    : UIntX;
  Vm       : T_BNX;
  Vn       : T_BNX;
  Vmn      : T_BNX;
  VR       : array [0..2] of T_BNX;
  VI       : array [0..2] of Integer;
  VI0      : Integer;
  VCompare : Integer;
begin
  VSign_m := m.Sign;
  VSign_n := n.Sign;

  m.Sign := 1;
  n.Sign := 1;
  try
    // (0, n) = n
    // (1, n) = 1
    // (m, 0) = m
    // (m, 1) = 1
    // (m, m) = m
    // (-m, n) = (m, n)

    if BN_IsInt (m, VUInt) then begin
      if VUInt = 0 then begin // m = 0
        BN_dup (R, n);        //  (0, n) = n
        Exit;
      end;
      if VUInt = 1 then begin   // m = 1
        BN_AssignInt (R, 1, 1); //  (1, n) = 1
        Exit;
      end;
    end;

    if BN_IsInt (n, VUInt) then begin
      if VUInt = 0 then begin // n = 0
        BN_dup (R, m);        //  (m, 0) = m
        Exit;
      end;
      if VUInt = 1 then begin   // n = 1
        BN_AssignInt (R, 1, 1); //  (m, 1) = 1
        Exit;
      end;
    end;

    VCompare := BN_DataCompare (m, n);

    if VCompare = 0 then begin
      BN_dup (R, m); //  (m, m) = m
      Exit;
    end;

    Vm_odd := BN_IsOdd (m);
    Vm_even := not Vm_odd;
    Vn_odd := BN_IsOdd (n);
    Vn_even := not Vn_odd;

    //  m, n ,  (m, n) = 2*(m/2, n/2)
    //  m , n ,  (m, n) = (m/2, n)
    //  n , m ,  (m, n) = (m, n/2)
    //  m, n   n > m,  (m, n) = (m, (n-m)/2) = ((n-m)/2, m)
    //  m, n   n < m,  (m, n) = ((m-n)/2, n)

    if Vm_even and Vn_even then begin
      BN_shr_(Vm, m, 1); // m div 2
      BN_shr_(Vn, n, 1); // n div 2
      BN_gcd_bin (R, Vm, Vn);
      BN_shl_(R, 1);     // * 2
      BN_Clear (Vm);
      BN_Clear (Vn);
      Exit;
    end;

    if Vm_even and Vn_odd then begin
      BN_shr_(Vm, m, 1); // m div 2
      BN_gcd_bin (R, Vm, n);
      BN_Clear (Vm);
      Exit;
    end;

    if Vm_odd and Vn_even then begin
      BN_shr_(Vn, n, 1); // n div 2
      BN_gcd_bin (R, m, Vn);
      BN_Clear (Vn);
      Exit;
    end;

    if Vm_odd and Vn_odd then begin
      if VCompare > 0 then begin
        // m > n
        BN_sub (Vmn, m, n); // (m - n)
        BN_shr_(Vmn, 1);    // (m - n) div 2
        BN_gcd_bin (R, Vmn, n);
      end else begin
        // m < n
        BN_sub (Vmn, n, m); // (n - m)
        BN_shr_(Vmn, 1);    // (n - m) div 2
        BN_gcd_bin (R, Vmn, m);
      end;
      BN_Clear (Vmn);
      Exit;
    end;

    if VCompare > 0 then begin
      // m > n
      BN_dup (VR[0], m);
      BN_dup (VR[1], n);
    end else begin
      // m < n
      BN_dup (VR[0], n);
      BN_dup (VR[1], m);
    end;

    VI[0] := 0;
    VI[1] := 1;
    VI[2] := 2;

    while true do begin
      BN_mod (VR[VI[2]], VR[VI[0]], VR[VI[1]]);
      if BN_IsZero (VR[VI[2]]) then begin
        BN_dup (R, VR[VI[1]]);
        Exit;
      end;

      VI0 := VI[0];
      VI[0] := VI[1];
      VI[1] := VI[2];
      VI[2] := VI0;
    end;
  finally
    m.Sign := VSign_m;
    n.Sign := VSign_n;

    BN_Clear (VR[0]);
    BN_Clear (VR[1]);
    BN_Clear (VR[2]);
  end;
end;

{
    
  Least common multiple

  (m,n)
  [m,n]
  lcm(m,n)

  lcm(m,n) = |m*n|/gcd(m,n)
}
procedure BN_lcm (var R: T_BNX; var m, n: T_BNX);
var
  Vm : T_BNX;
  Vd : T_BNX;
begin
  BN_mul (Vm, m, n);
  Vm.Sign := 1; // Vm = |m*n|
  BN_gcd (Vd, m, n);
  BN_div (R, Vm, Vd);
  BN_Clear (Vm);
  BN_Clear (Vd);
end;

{
  
  Factorial

  R = a * (a-1) * (a-2) * ... * 3 * 2 * 1
}
procedure BN_fact (var R: T_BNX; const a: T_BNX);
var
  UseR  : Boolean;
  VUInt : UIntX;
  VR    : T_BNX;
  sha   : T_BNX;
  o, Va : T_BNX;
  n, nn : UIntX;
  IsOne : Boolean;
begin
  if a.Sign < 0 then begin
    raise ERangeError.Create ('a! a<0');
  end;

  if BN_IsInt (a, VUInt) then begin
    BN_fact_int (R, VUInt);
    Exit;
  end;

  nn := BN_m2n_(a, sha);
  if nn = 0 then begin
    BN_dup (R, a);
  end else begin
    BN_dup (R, sha);
  end;
  BN_dup (o, a);

  UseR := true;
  while true do begin
    if UseR then begin
      // R, o -> VR, Va
      BN_sub_int (Va, o, 1, 1);
      n := BN_m2n_(Va, sha);
      if n = 0 then begin
        BN_mul (VR, R, Va);
      end else begin
        nn := nn + n;
        BN_mul (VR, R, sha);
      end;
      IsOne := BN_IsOne (Va);
    end else begin
      // VR, Va -> R, o
      BN_sub_int (o, Va, 1, 1);
      n := BN_m2n_(o, sha);
      if n = 0 then begin
        BN_mul (R, VR, o);
      end else begin
        nn := nn + n;
        BN_mul (R, VR, sha);
      end;
      IsOne := BN_IsOne (o);
    end;

    UseR := not UseR;
    if IsOne then Break;
  end;
  if UseR then begin
    BN_shl_(R, nn);
  end else begin
    BN_shl_(R, VR, nn);
  end;

  BN_Clear (VR);
  BN_Clear (sha);
  BN_Clear (o);
  BN_Clear (Va);
end;

{
  
  Factorial

  R = a * (a-1) * (a-2) * ... * 3 * 2 * 1
}
procedure BN_fact_int (var R: T_BNX; a: UIntX);
var
  UseR  : Boolean;
  VR    : T_BNX;
  sha   : UIntX;
  n, nn : UIntX;
begin
  if a <= 1 then begin
    BN_AssignInt (R, 1, 1); // 0! = 1 // 1! = 1
    Exit;
  end;

  nn := BN_m2n_int_(a, sha);
  if nn = 0 then begin
    BN_AssignInt (R, 1, a);
  end else begin
    BN_AssignInt (R, 1, sha);
  end;

  UseR := true;
  while true do begin
    if UseR then begin
      // R -> VR
      Dec (a);
      n := BN_m2n_int_(a, sha);
      if n = 0 then begin
        BN_mul_int (VR, R, 1, a);
      end else begin
        nn := nn + n;
        BN_mul_int (VR, R, 1, sha);
      end;
    end else begin
      // VR -> R
      Dec (a);
      n := BN_m2n_int_(a, sha);
      if n = 0 then begin
        BN_mul_int (R, VR, 1, a);
      end else begin
        nn := nn + n;
        BN_mul_int (R, VR, 1, sha);
      end;
    end;

    UseR := not UseR;
    if a = 1 then Break;
  end;
  if UseR then begin
    BN_shl_(R, nn);
  end else begin
    BN_shl_(R, VR, nn);
  end;

  BN_Clear (VR);
end;

end.

