unit mxBNMathProcs;

// Binuar 1.1
// Big number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{.$DEFINE BN_INLINE}

interface

uses
  Windows, mxBNMath;

function  BN_DataCount (const V: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_DataBytes (const V: T_BNX; var ACount: IntX): UIntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_DataBytes (const V: T_BNX): UIntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_DataNibbles (const V: T_BNX; var ACount: IntX): UIntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_DataNibbles (const V: T_BNX): UIntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_DataBits (const V: T_BNX; var ACount: IntX): UIntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_DataBits (const V: T_BNX): UIntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_DataValue (const V: T_BNX): UIntX; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_Clear (var V: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_Alloc (var V: T_BNX; ASign: ShortInt; ACount: IntX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF} overload;
function  BN_Alloc (var V: T_BNX; ACount: IntX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF} overload;
function  BN_Init (var V: T_BNX; ASign: ShortInt; ACount: IntX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_AssignZero (var V: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_AssignInt (var V: T_BNX; ASign: ShortInt; AUInt: UIntX): IntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_AssignSInt (var V: T_BNX; AInt: IntX): IntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_AssignBit (var V: T_BNX; ASign: ShortInt; ABit: UIntX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}

procedure BN_ClearBit (var V: T_BNX; ABit: UIntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
procedure BN_ClearBit (var V: T_BNX; const A: T_BNX; ABit: UIntX); overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_Pack (var V: T_BNX): IntX; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_Resize (var V: T_BNX; ACount: IntX): IntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_Resize (var V: T_BNX; ASign: ShortInt; ACount: IntX): IntX; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_Compare (const V1, V2: T_BNX): Integer; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_DataCompare (const V1, V2: T_BNX): Integer; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_DataCompare (const V1: T_BNX; V2: UIntX): Integer; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_IsZero (const V: T_BNX): Boolean; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_IsOne (const V: T_BNX): Boolean; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_IsInt (const V: T_BNX; var AUInt: UIntX): Boolean; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_IsInt (const V: T_BNX; var ASign: ShortInt; var AUInt: UIntX): Boolean; overload; {$IFDEF BN_INLINE} inline; {$ENDIF}

function  BN_IsOdd (const V: T_BNX): Boolean; {$IFDEF BN_INLINE} inline; {$ENDIF}
function  BN_IsEven (const V: T_BNX): Boolean; {$IFDEF BN_INLINE} inline; {$ENDIF}

implementation

uses
  mxBNMathUtils;

{
       
  The number of elements excluding the upper zeros
}
function BN_DataCount (const V: T_BNX): IntX;
{$IFDEF CPUX64}
// rcx | V
asm
     // mov     rax, [rcx].T_BNX.Count
        mov     rax, [rcx + T_BNX.Count]
        mov     rdx, [rcx + T_BNX.HIndex]
     // cmp     rdx, 0
     // jl      @r // HIndex < 0
        mov     rcx, [rcx + T_BNX.PHData]
        test    rcx, rcx
        jz      @r // PHData = nil
    @loop:
        mov     r8, [rcx]
        test    r8, r8
        jnz     @r
        dec     rax
        lea     rcx, [rcx - 8]
        dec     rdx
        jge     @loop
    @r:
end;
{$ELSE}
// eax | V
asm
        push    ebx

        mov     ecx, eax
     // mov     eax, [ecx].T_BNX.Count
        mov     eax, [ecx + T_BNX.Count]
        mov     edx, [ecx + T_BNX.HIndex]
     // cmp     edx, 0
     // jl      @r // HIndex < 0
        mov     ecx, [ecx + T_BNX.PHData]
        test    ecx, ecx
        jz      @r // PHData = nil
    @loop:
        mov     ebx, [ecx]
        test    ebx, ebx
        jnz     @r
        dec     eax
        lea     ecx, [ecx - 4]
        dec     edx
        jge     @loop
    @r:
        pop     ebx
end;
{$ENDIF}

{
       
  The number of elements excluding the upper zeros

  Pure pascal
}
function BN_DataCount_PP (const V: T_BNX): IntX;
var
  I : LoopInt;
begin
  Result := V.Count;
  for I := V.HIndex downto 0 do begin
    if V.Data[I] <> 0 then Break;
    Dec (Result);
  end;
end;

{
       
  The number of bytes excluding the upper zeros
}
function BN_DataBytes (const V: T_BNX; var ACount: IntX): UIntX;
var
  VUCount : UIntX;
begin
  Result := 0;
  ACount := 0;
  if V.Sign = 0 then Exit;

  ACount := BN_DataCount (V);
  if ACount <= 0 then Exit;

  VUCount := ACount - 1;
  Result := BN_Bytes (V.Data[VUCount]);

  {$IFDEF CPUX64}
  Result := Result + (VUCount shl 3); // ... Result + VUCount * 8;
  {$ELSE}
  Result := Result + (VUCount shl 2); // ... Result + VUCount * 4;
  {$ENDIF}
end;

{
       
  The number of bytes excluding the upper zeros
}
function BN_DataBytes (const V: T_BNX): UIntX;
var
  VCount : IntX;
begin
  Result := BN_DataBytes (V, VCount);
end;

{
       
  The number of half-bytes excluding the upper zeros
}
function BN_DataNibbles (const V: T_BNX; var ACount: IntX): UIntX;
var
  VUCount : UIntX;
begin
  Result := 0;
  ACount := 0;
  if V.Sign = 0 then Exit;

  ACount := BN_DataCount (V);
  if ACount <= 0 then Exit;

  VUCount := ACount - 1;
  Result := BN_Nibbles (V.Data[VUCount]);

  {$IFDEF CPUX64}
  Result := Result + (VUCount shl 4); // ... Result + VUCount * 16;
  {$ELSE}
  Result := Result + (VUCount shl 3); // ... Result + VUCount * 8;
  {$ENDIF}
end;

{
       
  The number of half-bytes excluding the upper zeros
}
function BN_DataNibbles (const V: T_BNX): UIntX;
var
  VCount : IntX;
begin
  Result := BN_DataNibbles (V, VCount);
end;

{
       
  The number of bits excluding the upper zeros
}
function BN_DataBits (const V: T_BNX; var ACount: IntX): UIntX;
var
  VUCount : UIntX;
begin
  Result := 0;
  ACount := 0;
  if V.Sign = 0 then Exit;

  ACount := BN_DataCount (V);
  if ACount <= 0 then Exit;

  VUCount := ACount - 1;
  Result := BN_Bits (V.Data[VUCount]);

  {$IFDEF CPUX64}
  Result := Result + (VUCount shl 6); // ... Result + VUCount * 64;
  {$ELSE}
  Result := Result + (VUCount shl 5); // ... Result + VUCount * 32;
  {$ENDIF}
end;

{
       
  The number of bits excluding the upper zeros
}
function BN_DataBits (const V: T_BNX): UIntX;
var
  VCount : IntX;
begin
  Result := BN_DataBits (V, VCount);
end;

{
       0
  Return the value of a single element or 0
}
function BN_DataValue (const V: T_BNX): UIntX;
var
  VCount : IntX;
begin
  Result := 0;

  VCount := BN_DataCount (V);
  if VCount = 1 then Result := V.Data[0];
end;

{
}
function BN_Clear (var V: T_BNX): IntX;
begin
  Result := 0;

  SetLength (V.Data, 0);

  V.Sign := 0;
  V.Count := 0;
  V.HIndex := -1;
  V.PLData := nil;
  V.PHData := nil;
end;

{
}
function BN_Alloc (var V: T_BNX; ASign: ShortInt; ACount: IntX): IntX;
var
  VHIndex : IntX;
begin
  if (ASign = 0) or (ACount <= 0) then begin
    Result := BN_Clear (V);
    Exit;
  end;

  Result := ACount;
  V.Sign := ASign;

  VHIndex := ACount - 1;
  if HIGH(V.Data) = VHIndex then Exit;

  SetLength (V.Data, ACount);

  V.Count := ACount;
  V.HIndex := VHIndex;
  V.PLData := @V.Data[0];
  V.PHData := @V.Data[VHIndex];
end;

{
}
function BN_Alloc (var V: T_BNX; ACount: IntX): IntX;
var
  VHIndex : IntX;
begin
  if (ACount <= 0) then begin
    Result := BN_Clear (V);
    Exit;
  end;

  Result := ACount;
  // V.Sign := ...

  VHIndex := ACount - 1;
  if HIGH(V.Data) = VHIndex then Exit;

  SetLength (V.Data, ACount);

  V.Count := ACount;
  V.HIndex := VHIndex;
  V.PLData := @V.Data[0];
  V.PHData := @V.Data[VHIndex];
end;

{
}
function BN_Init (var V: T_BNX; ASign: ShortInt; ACount: IntX): IntX;
begin
  Result := BN_Alloc (V, ASign, ACount);
  if V.Count <= 0 then Exit;

  BN_ZeroData (V.PLData, V.Count);
end;

{
}
function BN_AssignZero (var V: T_BNX): IntX;
begin
  Result := BN_Alloc (V, 0, 0);
end;

{
}
function BN_AssignInt (var V: T_BNX; ASign: ShortInt; AUInt: UIntX): IntX;
begin
  if AUInt = 0 then begin
    Result := BN_AssignZero (V);
    Exit;
  end;

  Result := BN_Alloc (V, ASign, 1);
  if Result > 0 then V.Data[0] := AUInt;
end;

{
  AInt - signed integer
}
function BN_AssignSInt (var V: T_BNX; AInt: IntX): IntX;
var
  VSign : ShortInt;
begin
  if AInt = 0 then begin
    Result := BN_AssignZero (V);
    Exit;
  end;

  if AInt > 0 then VSign := 1
  else begin
    VSign := -1;
    AInt := -AInt;
  end;

  Result := BN_Alloc (V, VSign, 1);
  if Result > 0 then V.Data[0] := AInt;
end;

{
          ABit
  Create a number with the only bit set at the ABit position

  ABit = 0,1,...
}
function BN_AssignBit (var V: T_BNX; ASign: ShortInt; ABit: UIntX): IntX;
const
  {$IFDEF CPUX64}
  CBits = 64;
  {$ELSE}
  CBits = 32;
  {$ENDIF}
var
  VmBits : UIntX;
  VCount : IntX;
begin
  (*
  VCount := cdiv_mod (ABit, CBits, VmBits);
  if VCount <= 0 then VCount := 1;
  // *)
  VCount := div_mod (ABit, CBits, VmBits) + 1;

  Result := BN_Init (V, ASign, VCount);
  V.PHData^ := UIntX(1) shl VmBits;
end;

{
      ABit
  Clear the bit at the ABit position

  ABit = 0,1,...
}
procedure BN_ClearBit (var V: T_BNX; ABit: UIntX);
const
  {$IFDEF CPUX64}
  CBits = 64;
  {$ELSE}
  CBits = 32;
  {$ENDIF}
var
  VmBits : UIntX;
  VCount : IntX;
  VIndex : IntX;
  Vmask  : UIntX;
begin
  (*
  VCount := cdiv_mod (ABit, CBits, VmBits);
  if VCount <= 0 then VCount := 1;
  // *)
  VCount := div_mod (ABit, CBits, VmBits) + 1;

  if VCount > BN_DataCount (V) then Exit;

  VIndex := VCount - 1;
  Vmask := not (UIntX(1) shl VmBits);
  V.Data[VIndex] := V.Data[VIndex] and Vmask;
  BN_Pack (V);
end;

{
      ABit
  Clear the bit at the ABit position

  ABit = 0,1,...
}
procedure BN_ClearBit (var V: T_BNX; const A: T_BNX; ABit: UIntX);
const
  {$IFDEF CPUX64}
  CBits = 64;
  {$ELSE}
  CBits = 32;
  {$ENDIF}
var
  VmBits : UIntX;
  VCount : IntX;
  VIndex : IntX;
  Vmask  : UIntX;
begin
  BN_dup (V, A);

  (*
  VCount := cdiv_mod (ABit, CBits, VmBits);
  if VCount <= 0 then VCount := 1;
  // *)
  VCount := div_mod (ABit, CBits, VmBits) + 1;

  if VCount > BN_DataCount (V) then Exit;

  VIndex := VCount - 1;
  Vmask := not (UIntX(1) shl VmBits);
  V.Data[VIndex] := V.Data[VIndex] and Vmask;
  BN_Pack (V);
end;

{
    
  Remove the upper zeros
}
function BN_Pack (var V: T_BNX): IntX;
begin
  Result := BN_DataCount (V);
  if Result = V.Count then Exit;

  BN_Alloc (V, Result);
end;

{
}
function BN_Resize (var V: T_BNX; ACount: IntX): IntX;
var
  VCount  : IntX;
  VDCount : IntX;
begin
  VCount := V.Count;
  if ACount = VCount then begin
    Result := ACount;
    Exit;
  end;

  Result := BN_Alloc (V, ACount);

  VDCount := V.Count - VCount;
  if VDCount <= 0 then Exit;

  BN_ZeroData (@V.Data[VCount], VDCount);
end;

{
}
function BN_Resize (var V: T_BNX; ASign: ShortInt; ACount: IntX): IntX;
var
  VCount  : IntX;
  VDCount : IntX;
begin
  V.Sign := ASign;

  VCount := V.Count;
  if ACount = VCount then begin
    Result := ACount;
    Exit;
  end;

  Result := BN_Alloc (V, ACount);

  VDCount := V.Count - VCount;
  if VDCount <= 0 then Exit;

  BN_ZeroData (@V.Data[VCount], VDCount);
end;

{
   
  Signed comparison

  Result
  V1 = V2 :  0
  V1 > V2 : +1
  V1 < V2 : -1
}
function BN_Compare (const V1, V2: T_BNX): Integer;
var
  VSign1 : IntX;
  VSign2 : IntX;
begin
  VSign1 := V1.Sign;
  VSign2 := V2.Sign;

  if (VSign1 = 0) then begin
    // V1 = 0
    // 0 =  0 |  0
    // 0 < +1 | -1
    // 0 > -1 | +1
    Result := -VSign2;
    Exit;
  end;

  if (VSign2 = 0) then begin
    // V2 = 0
    //  0 = 0 |  0
    // +1 > 0 | +1
    // -1 < 0 | -1
    Result := VSign1;
    Exit;
  end;

  Result := BN_DataCompare (V1, V2);

  if Result > 0 then begin
    // |V1| > |V2|
    // +2 > +1 | +1
    // +2 > -1 | +1
    // -2 < +1 | -1
    // -2 < -1 | -1
    Result := VSign1;
    Exit;
  end;

  if Result < 0 then begin
    // |V1| < |V2|
    // +1 < +2 | -1
    // -1 < +2 | -1
    // +1 > -2 | +1
    // -1 > -2 | +1
    Result := -VSign2;
    Exit;
  end;

  // |V1| = |V2|
  // +2 = +2 |  0
  // +2 > -2 | +1
  // -2 < +2 | -1
  // -2 = -2 |  0
  if VSign1 <> VSign2 then begin
    Result := VSign1;
  end;
end;

{
}
function BNX_DataCompare (const V1, V2; Count: IntX): Integer;
{$IFDEF CPUX64}
// rcx | V1
// rdx | V2
// r8  | Count
asm
        xor     rax, rax

        dec     r8
        jl      @r // Count <= 0
    @loop:
        mov     r9, [rcx]
        cmp     r9, [rdx]
        ja      @a
        jb      @b
        lea     rcx, [rcx - 8]
        lea     rdx, [rdx - 8]
        dec     r8
        jge     @loop
        jmp     @r
    @a: inc     rax
        jmp     @r
    @b: dec     rax
    @r:
end;
{$ELSE}
// eax | V1    | ebx
// edx | V2    |
// ecx | Count |
asm
        push    ebx
        push    esi

        mov     ebx, eax
        xor     eax, eax

        dec     ecx
        jl      @r // Count <= 0
    @loop:
        mov     esi, [ebx]
        cmp     esi, [edx]
        ja      @a
        jb      @b
        lea     ebx, [ebx - 4]
        lea     edx, [edx - 4]
        dec     ecx
        jge     @loop
        jmp     @r
    @a: inc     eax
        jmp     @r
    @b: dec     eax
    @r:
        pop     esi
        pop     ebx
end;
{$ENDIF}

{
   
  Unsigned comparison

  Result
  V1.Data = V2.Data :  0
  V1.Data > V2.Data : +1
  V1.Data < V2.Data : -1
}
function BN_DataCompare (const V1, V2: T_BNX): Integer;
var
  I       : IntX;
  VCount1 : IntX;
  VCount2 : IntX;
begin
  VCount1 := BN_DataCount (V1);
  VCount2 := BN_DataCount (V2);

  if VCount1 > VCount2 then begin
    Result := +1;
    Exit;
  end;

  if VCount1 < VCount2 then begin
    Result := -1;
    Exit;
  end;

  // VCount1 = VCount2

  I := VCount1 - 1;
  Result := BNX_DataCompare (V1.Data[I], V2.Data[I], VCount1);
end;

{
   
  Unsigned comparison

  Result
  V1.Data = V2.Data :  0
  V1.Data > V2.Data : +1
  V1.Data < V2.Data : -1
}
function BN_DataCompare_(const V1, V2: T_BNX): Integer;
var
  I       : LoopInt;
  VCount1 : IntX;
  VCount2 : IntX;
  VValue1 : UIntX;
  VValue2 : UIntX;
begin
  Result := 0;

  VCount1 := BN_DataCount (V1);
  VCount2 := BN_DataCount (V2);

  if VCount1 > VCount2 then begin
    Result := +1;
    Exit;
  end;

  if VCount1 < VCount2 then begin
    Result := -1;
    Exit;
  end;

  // VCount1 = VCount2

  for I := VCount1 - 1 downto 0 do begin
    VValue1 := V1.Data[I];
    VValue2 := V2.Data[I];
    if VValue1 > VValue2 then begin
      Result := +1;
      Exit;
    end;
    if VValue1 < VValue2 then begin
      Result := -1;
      Exit;
    end;
  end;
end;

{
   
  Unsigned comparison

  Result
  V1.Data = V2 :  0
  V1.Data > V2 : +1
  V1.Data < V2 : -1
}
function BN_DataCompare (const V1: T_BNX; V2: UIntX): Integer;
var
  VCount1 : IntX;
  VValue1 : UIntX;
begin
  Result := 0;

  VCount1 := BN_DataCount (V1);

  if VCount1 <= 0 then begin
    // |V1|  = 0
    // |V2|  = 0 | 0
    // |V2| != 0 | -1
    if V2 <> 0 then Result := -1;
    Exit;
  end;

  // |V1| != 0

  if (VCount1 > 1) or (V2 = 0) then begin
    Result := +1;
    Exit;
  end;

  // VCount1 = 1

  VValue1 := V1.Data[0];
  if VValue1 > V2 then begin
    Result := +1;
    Exit;
  end;
  if VValue1 < V2 then begin
    Result := -1;
    Exit;
  end;
end;

{
}
function BN_IsZero (const V: T_BNX): Boolean;
var
  VCount : IntX;
begin
  VCount := BN_DataCount (V);
  Result := (V.Sign = 0) or (VCount = 0);
end;

{
}
function BN_IsOne (const V: T_BNX): Boolean;
var
  VCount : IntX;
begin
  Result := false;
  VCount := BN_DataCount (V);
  if (V.Sign <> 0) and (VCount = 1) then Result := V.Data[0] = 1;
end;

{
}
function BN_IsInt (const V: T_BNX; var AUInt: UIntX): Boolean;
var
  VCount : IntX;
begin
  Result := false;
  AUInt := 0;

  VCount := BN_DataCount (V);
  if VCount > 1 then Exit;

  Result := true;
  if (V.Sign <> 0) and (VCount = 1) then begin
    AUInt := V.Data[0];
  end;
end;

{
}
function BN_IsInt (const V: T_BNX; var ASign: ShortInt; var AUInt: UIntX): Boolean;
var
  VCount : IntX;
begin
  Result := false;
  ASign := 0;
  AUInt := 0;

  VCount := BN_DataCount (V);
  if VCount > 1 then Exit;

  Result := true;
  if (V.Sign <> 0) and (VCount = 1) then begin
    ASign := V.Sign;
    AUInt := V.Data[0];
  end;
end;

{
  ?  
  ? Odd number
}
function BN_IsOdd (const V: T_BNX): Boolean;
var
  VCount : IntX;
begin
  VCount := BN_DataCount (V);
  if VCount <= 0 then begin
    Result := false; // 0 is even
    Exit;
  end;

  Result := Odd (V.Data[0]);
end;

{
  ? ׸ 
  ? Even number
}
function BN_IsEven (const V: T_BNX): Boolean;
var
  VCount : IntX;
begin
  VCount := BN_DataCount (V);
  if VCount <= 0 then begin
    Result := true; // 0 is even
    Exit;
  end;

  Result := not Odd (V.Data[0]);
end;

end.

