unit mxBNMathUtils;

// Binuar 1.1
// Big number arithmetic
// ---
// Copyright  2000-3000, Andrey A. Meshkov (AL-CHEMIST)
// All rights reserved
//
// http://maalchemist.ru
// maa@maalchemist.ru
// maalchemist@yandex.ru
// maalchemist@gmail.com
//
// Free for any use.
// No warranty of any kind.
// ---

{$I version.inc}
{.$DEFINE BN_INLINE}

interface

uses
  Windows, mxBNMath;

function  cdiv (a, b: UIntX): UIntX; overload;
function  cdiv (a, b: IntX): IntX; overload;
function  cdiv_mod (a, b: UIntX; var a_mod: UIntX): UIntX; overload;
function  cdiv_mod (a, b: IntX; var a_mod: IntX): IntX; overload;
function  div_mod (a, b: UIntX; var a_mod: UIntX): UIntX; overload;
function  div_mod (a, b: IntX; var a_mod: IntX): IntX; overload;

function  BN_Bytes (AValue: UIntX): UIntX;
function  BN_Nibbles (AValue: UIntX): UIntX;
function  BN_Bits (AValue: UIntX): UIntX;

function  BN_Size (ACount: IntX): UIntX; overload;
function  BN_Size (ACount: UIntX): UIntX; overload;

function  BN_IncPointer (P: Pointer): Pointer;
function  BN_IncPointer_(var P): Pointer;
function  BN_DecPointer (P: Pointer): Pointer;
function  BN_DecPointer_(var P): Pointer;

procedure BN_ZeroData (DST: Pointer; Count: IntX);
procedure BN_ZeroDataDesc (DST: Pointer; Count: IntX);
procedure BN_FillData (DST: Pointer; Count: IntX; V: UIntX);
procedure BN_FillDataDesc (DST: Pointer; Count: IntX; V: UIntX);
procedure BN_CopyData (DST, SRC: Pointer; Count: IntX);
procedure BN_CopyDataDesc (DST, SRC: Pointer; Count: IntX);

implementation

uses
  SysUtils;

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
}
function cdiv (a, b: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a |
// rdx | b | r10
asm
        mov     r10, rdx
        mov     rax, rcx
        xor     rdx, rdx
        div     r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
end;
{$ELSE}
// eax | a |
// edx | b | ecx
asm
        mov     ecx, edx
        xor     edx, edx
        div     ecx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
end;
{$ENDIF}

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
}
function cdiv (a, b: IntX): IntX;
{$IFDEF CPUX64}
// rcx | a |
// rdx | b | r10
asm
        mov     r10, rdx
        mov     rax, rcx
        cqo
        idiv    r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
end;
{$ELSE}
// eax | a |
// edx | b | ecx
asm
        mov     ecx, edx
        cdq
        idiv    ecx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
end;
{$ENDIF}

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
                 a_mod  = a mod b
}
function cdiv_mod (a, b: UIntX; var a_mod: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
        mov     r10, rdx
        mov     rax, rcx
        xor     rdx, rdx
        div     r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
        mov     [r8], rdx
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        xor     edx, edx
        div     ebx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
        mov     [ecx], edx

        pop     ebx
end;
{$ENDIF}

{
  a mod b == 0 : result = a div b
  a mod b != 0 : result = a div b + 1
                 a_mod  = a mod b
}
function cdiv_mod (a, b: IntX; var a_mod: IntX): IntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
        mov     r10, rdx
        mov     rax, rcx
        cqo
        idiv    r10      // rdx = a mod b
        test    rdx, rdx // ? a mod b == 0
        jz      @r
        inc     rax      // a mod b != 0
  @r:
        mov     [r8], rdx
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        cdq
        idiv    ebx      // edx = a mod b
        test    edx, edx // ? a mod b == 0
        jz      @r
        inc     eax      // a mod b != 0
  @r:
        mov     [ecx], edx

        pop     ebx
end;
{$ENDIF}

{
  result = a div b
  a_mod  = a mod b
}
function div_mod (a, b: UIntX; var a_mod: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
        mov     r10, rdx
        mov     rax, rcx
        xor     rdx, rdx
        div     r10
        mov     [r8], rdx // rdx = a mod b
  @r:
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        xor     edx, edx
        div     ebx
        mov     [ecx], edx // edx = a mod b
  @r:
        pop     ebx
end;
{$ENDIF}

{
  result = a div b
  a_mod  = a mod b
}
function div_mod (a, b: IntX; var a_mod: IntX): IntX;
{$IFDEF CPUX64}
// rcx | a      |
// rdx | b      | r10
// r8  | @a_mod |
asm
        mov     r10, rdx
        mov     rax, rcx
        cqo
        idiv    r10
        mov     [r8], rdx // rdx = a mod b
  @r:
end;
{$ELSE}
// eax | a      |
// edx | b      | ebx
// ecx | @a_mod |
asm
        push    ebx

        mov     ebx, edx
        cdq
        idiv    ebx
        mov     [ecx], edx // edx = a mod b
  @r:
        pop     ebx
end;
{$ENDIF}

{
}
function BN_Size (ACount: IntX): UIntX;
{$IFDEF CPUX64}
// rcx | ACount
asm
        xor     rax, rax
        cmp     rcx, 0
        jle     @r
        lea     rax, [rcx * 8]
  @r:
end;
{$ELSE}
// eax | ACount | ecx
asm
        mov     ecx, eax
        xor     eax, eax
        cmp     ecx, 0
        jle     @r
        lea     eax, [ecx * 4]
  @r:
end;
{$ENDIF}

{
       
  The number of bytes excluding the upper zeros
}
function BN_Bytes (AValue: UIntX): UIntX;
var
  VBits : UIntX;
begin
  VBits := BN_Bits (AValue);
  Result := cdiv (VBits, 8);
end;

{
       
  The number of half-bytes excluding the upper zeros
}
function BN_Nibbles (AValue: UIntX): UIntX;
var
  VBits : UIntX;
begin
  VBits := BN_Bits (AValue);
  Result := cdiv (VBits, 4);
end;

{
       
   shr

  The number of bits excluding the upper zeros
  Using shr function
}
function BN_Bits (AValue: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | AValue
asm
        xor     rax, rax

        test    rcx, rcx
        jz      @r
  @a:
        inc     rax
        shr     rcx, 1
        test    rcx, rcx
        jnz     @a
  @r:
end;
{$ELSE}
// eax | AValue | ecx
asm
        mov     ecx, eax
        xor     eax, eax

        test    ecx, ecx
        jz      @r
  @a:
        inc     eax
        shr     ecx, 1
        test    ecx, ecx
        jnz     @a
  @r:
end;
{$ENDIF}

{
       
   shr

  The number of bits excluding the upper zeros
  Using shr function

  Pure pascal
}
function BN_Bits_PP (AValue: UIntX): UIntX;
begin
  Result := 0;
  while AValue <> 0 do begin
    Inc (Result);
    AValue := AValue shr 1;
  end;
end;

{
       
   shl

  The number of bits excluding the upper zeros
  Using shl function
}
function BN_Bits_SHL (AValue: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | AValue
asm
        test    rcx, rcx
        jnz     @a
        xor     rax, rax
        jmp     @r

  @a:
        mov     rax, 64

        shl     rcx, 1
        jc      @r
  @b:
        dec     rax
        shl     rcx, 1
        jnc     @b
  @r:
end;
{$ELSE}
// eax | AValue | ecx
asm
        test    eax, eax
        jz      @r

  @a:
        mov     ecx, eax
        mov     eax, 32

        shl     ecx, 1
        jc      @r
  @b:
        dec     eax
        shl     ecx, 1
        jnc     @b
  @r:
end;
{$ENDIF}

{
}
function BN_Size (ACount: UIntX): UIntX;
{$IFDEF CPUX64}
// rcx | ACount | rax
asm
        mov     rax, rcx
        test    rax, rax
        jz      @r
        lea     rax, [rax * 8]
  @r:
end;
{$ELSE}
// eax | ACount | ecx
asm
        test    eax, eax
        jz      @r
        lea     eax, [eax * 4]
  @r:
end;
{$ENDIF}

{
  Pure pascal
}
function BN_Size_PP (ACount: IntX): UIntX;
begin
  if ACount <= 0 then begin
    Result := 0;
    Exit;
  end;

  {$IFDEF CPUX64}
  Result := ACount shl 3; // ACount * 8;
  {$ELSE}
  Result := ACount shl 2; // ACount * 4;
  {$ENDIF}
end;

{
}
function BN_IncPointer (P: Pointer): Pointer;
{$IFDEF CPUX64}
// rcx | P
asm
        lea     rax, [rcx + 8]
end;
{$ELSE}
// eax | P
asm
        lea     eax, [eax + 4]
end;
{$ENDIF}

{
}
function BN_IncPointer_(var P): Pointer;
{$IFDEF CPUX64}
// rcx | @P
asm
        mov     rax, [rcx]
        lea     rax, [rax + 8]
        mov     [rcx], rax
end;
{$ELSE}
// eax | @P | ecx
asm
        mov     ecx, eax
        mov     eax, [ecx]
        lea     eax, [eax + 4]
        mov     [ecx], eax
end;
{$ENDIF}

{
}
function BN_DecPointer (P: Pointer): Pointer;
{$IFDEF CPUX64}
// rcx | P
asm
        lea     rax, [rcx - 8]
end;
{$ELSE}
// eax | P
asm
        lea     eax, [eax - 4]
end;
{$ENDIF}

{
}
function BN_DecPointer_(var P): Pointer;
{$IFDEF CPUX64}
// rcx | @P
asm
        mov     rax, [rcx]
        lea     rax, [rax - 8]
        mov     [rcx], rax
end;
{$ELSE}
// eax | @P | ecx
asm
        mov     ecx, eax
        mov     eax, [ecx]
        lea     eax, [eax - 4]
        mov     [ecx], eax
end;
{$ENDIF}

{
  ascending address order
}
procedure BN_ZeroData (DST: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST   |
// rdx | Count |
//     | 0     | rax
asm
        dec     rdx
        jl      @r // Count <= 0
        xor     rax, rax
  @loop:
        mov     [rcx], rax // DST = 0
        lea     rcx, [rcx + 8]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST   |
// edx | Count |
//     | 0     | ecx
asm
        dec     edx
        jl      @r // Count <= 0
        xor     ecx, ecx
  @loop:
        mov     [eax], ecx // DST = 0
        lea     eax, [eax + 4]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
  descending address order
}
procedure BN_ZeroDataDesc (DST: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST   |
// rdx | Count |
//     | 0     | rax
asm
        dec     rdx
        jl      @r // Count <= 0
        xor     rax, rax
  @loop:
        mov     [rcx], rax // DST = 0
        lea     rcx, [rcx - 8]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST   |
// edx | Count |
//     | 0     | ecx
asm
        dec     edx
        jl      @r // Count <= 0
        xor     ecx, ecx
  @loop:
        mov     [eax], ecx // DST = 0
        lea     eax, [eax - 4]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
  ascending address order
}
procedure BN_FillData (DST: Pointer; Count: IntX; V: UIntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | Count
// r8  | V
asm
        dec     rdx
        jl      @r // Count <= 0
  @loop:
        mov     [rcx], r8 // DST = V
        lea     rcx, [rcx + 8]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | Count
// ecx | V
asm
        dec     edx
        jl      @r // Count <= 0
  @loop:
        mov     [eax], ecx // DST = V
        lea     eax, [eax + 4]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
  descending address order
}
procedure BN_FillDataDesc (DST: Pointer; Count: IntX; V: UIntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | Count
// r8  | V
asm
        dec     rdx
        jl      @r // Count <= 0
  @loop:
        mov     [rcx], r8 // DST = V
        lea     rcx, [rcx - 8]
        dec     rdx
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | Count
// ecx | V
asm
        dec     edx
        jl      @r // Count <= 0
  @loop:
        mov     [eax], ecx // DST = V
        lea     eax, [eax - 4]
        dec     edx
        jge     @loop
  @r:
end;
{$ENDIF}

{
  ascending address order
}
procedure BN_CopyData (DST, SRC: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | SRC
// r8  | Count
asm
        dec     r8
        jl      @r // Count <= 0
  @loop:
        mov     rax, [rdx] // rax = SRC
        mov     [rcx], rax // DST = SRC
        lea     rcx, [rcx + 8]
        lea     rdx, [rdx + 8]
        dec     r8
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | SRC
// ecx | Count
asm
        push    ebx

        dec     ecx
        jl      @r // Count <= 0
  @loop:
        mov     ebx, [edx] // ebx = SRC
        mov     [eax], ebx // DST = SRC
        lea     eax, [eax + 4]
        lea     edx, [edx + 4]
        dec     ecx
        jge     @loop
  @r:
        pop     ebx
end;
{$ENDIF}

{
  descending address order
}
procedure BN_CopyDataDesc (DST, SRC: Pointer; Count: IntX);
{$IFDEF CPUX64}
// rcx | DST
// rdx | SRC
// r8  | Count
asm
        dec     r8
        jl      @r // Count <= 0
  @loop:
        mov     rax, [rdx] // rax = SRC
        mov     [rcx], rax // DST = SRC
        lea     rcx, [rcx - 8]
        lea     rdx, [rdx - 8]
        dec     r8
        jge     @loop
  @r:
end;
{$ELSE}
// eax | DST
// edx | SRC
// ecx | Count
asm
        push    ebx

        dec     ecx
        jl      @r // Count <= 0
  @loop:
        mov     ebx, [edx] // ebx = SRC
        mov     [eax], ebx // DST = SRC
        lea     eax, [eax - 4]
        lea     edx, [edx - 4]
        dec     ecx
        jge     @loop
  @r:
        pop     ebx
end;
{$ENDIF}

end.

